from __future__ import annotations

from pydantic import BaseModel


class Broker(BaseModel):
    pass


class MQTTBroker(Broker):
    host: str = "localhost"
    port: int = 1883
    username: str = ""
    password: str = ""


class AMQPBroker(Broker):
    vhost: str = "/"
    topicExchange: str = "amq.topic"
    host: str = "localhost"
    port: int = 5672
    username: str = "guest"
    password: str = "guest"


class RedisBroker(Broker):
    db: int = 0
    host: str = "localhost"
    port: int = 6379
    username: str = ""
    password: str = ""



from __future__ import annotations

import math
from dataclasses import dataclass
from typing import Any


@dataclass
class Point:
    x: float | None = 0.0
    y: float | None = 0.0
    z: float | None = 0.0

    def __sub__(self, other: Any):
        if isinstance(other, Point):
            return Point(self.x - other.x, self.y - other.y, self.z - other.z)
        elif isinstance(other, (float, int)):
            return Point(self.x - other, self.y - other, self.z - other)
        else:
            raise ValueError(
                f"Cannot perform substraction Point with type {type(other)}"
            )

    def __eq__(self, other):
        if isinstance(other, Point):
            return self.x == other.x and self.y == other.y and self.z == other.z
        if isinstance(other, dict):
            return (
                self.x == other["x"] and self.y == other["y"] and self.z == other["z"]
            )
        return False

    def __lt__(self, other):
        if isinstance(other, Point):
            return (self.x, self.y, self.z) < (other.x, other.y, other.z)
        elif isinstance(other, (dict)):
            return (self.x, self.y, self.z) < (other["x"], other["y"], other["z"])
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, Point):
            return (self.x, self.y, self.z) <= (other.x, other.y, other.z)
        elif isinstance(other, (dict)):
            return (self.x, self.y, self.z) <= (other["x"], other["y"], other["z"])
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, Point):
            return (self.x, self.y, self.z) > (other.x, other.y, other.z)
        elif isinstance(other, (dict)):
            return (self.x, self.y, self.z) > (other["x"], other["y"], other["z"])
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, Point):
            return (self.x, self.y, self.z) >= (other.x, other.y, other.z)
        elif isinstance(other, (dict)):
            return (self.x, self.y, self.z) >= (other["x"], other["y"], other["z"])
        return NotImplemented

    def abs(self):
        return math.sqrt(self.x**2 + self.y**2 + self.z**2)

    def __add__(self, other):
        if isinstance(other, Point):
            return Point(self.x + other.x, self.y + other.y, self.z + other.z)
        elif isinstance(other, (float, int)):
            return Point(self.x + other, self.y + other, self.z + other)
        else:
            raise ValueError(
                f"Cannot perform addition of Point with type {type(other)}"
            )


@dataclass
class Orientation:
    roll: float | None = 0.0
    pitch: float | None = 0.0
    yaw: float | None = 0.0

    def __init__(
        self,
        roll: float | None = 0.0,
        pitch: float | None = 0.0,
        yaw: float | None = 0.0,
        *,
        x: float | None = None,
        y: float | None = None,
        z: float | None = None,
    ) -> None:
        # Accept x/y/z as aliases for roll/pitch/yaw (robotics convention).
        # If both are provided, the explicit roll/pitch/yaw take precedence
        # only when they differ from the default (0.0).
        self.roll = x if x is not None and roll == 0.0 else roll
        self.pitch = y if y is not None and pitch == 0.0 else pitch
        self.yaw = z if z is not None and yaw == 0.0 else yaw

    def __sub__(self, other):
        if isinstance(other, Orientation):
            return Orientation(
                self.roll - other.roll, self.pitch - other.pitch, self.yaw - other.yaw
            )
        elif isinstance(other, (float, int)):
            return Orientation(self.roll - other, self.pitch - other, self.yaw - other)
        else:
            raise ValueError(
                f"Cannot perform substraction Point with type {type(other)}"
            )

    def __eq__(self, other):
        if isinstance(other, Orientation):
            return (
                self.roll == other.roll
                and self.pitch == other.pitch
                and self.yaw == other.yaw
            )
        elif isinstance(other, (dict)):
            return (
                self.roll == other["roll"]
                and self.pitch == other["pitch"]
                and self.yaw == other["yaw"]
            )
        return False

    def abs(self):
        return math.sqrt(self.roll**2 + self.pitch**2 + self.yaw**2)

    def __add__(self, other):
        if isinstance(other, Orientation):
            return Orientation(
                self.roll + other.roll, self.pitch + other.pitch, self.yaw + other.yaw
            )
        elif isinstance(other, (float, int)):
            return Orientation(self.roll + other, self.pitch + other, self.yaw + other)
        else:
            raise ValueError(
                f"Cannot perform addition of Point with type {type(other)}"
            )

    def __gt__(self, other):
        if isinstance(other, Orientation):
            return (self.roll, self.pitch, self.yaw) > (
                other.roll,
                other.pitch,
                other.yaw,
            )
        elif isinstance(other, (dict)):
            return (self.roll, self.pitch, self.yaw) > (
                other["roll"],
                other["pitch"],
                other["yaw"],
            )
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, Orientation):
            return (self.roll, self.pitch, self.yaw) >= (
                other.roll,
                other.pitch,
                other.yaw,
            )
        elif isinstance(other, (dict)):
            return (self.roll, self.pitch, self.yaw) >= (
                other["roll"],
                other["pitch"],
                other["yaw"],
            )
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, Orientation):
            return (self.roll, self.pitch, self.yaw) < (
                other.roll,
                other.pitch,
                other.yaw,
            )
        elif isinstance(other, (dict)):
            return (self.roll, self.pitch, self.yaw) < (
                other["roll"],
                other["pitch"],
                other["yaw"],
            )
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, Orientation):
            return (self.roll, self.pitch, self.yaw) <= (
                other.roll,
                other.pitch,
                other.yaw,
            )
        elif isinstance(other, (dict)):
            return (self.roll, self.pitch, self.yaw) <= (
                other["roll"],
                other["pitch"],
                other["yaw"],
            )


@dataclass
class Pose:
    translation: Point
    orientation: Orientation



from __future__ import annotations

from unittest.mock import patch

from goalee.goal import Goal, GoalState
from goalee.repeater import GoalRepeater


class MockGoal(Goal):
    def __init__(self, complete_on_enter=True, fail_on_enter=False, **kwargs):
        super().__init__(**kwargs)
        self._complete_on_enter = complete_on_enter
        self._fail_on_enter = fail_on_enter

    def on_enter(self):
        pass

    def tick(self):
        if self._complete_on_enter:
            self.set_state(GoalState.COMPLETED)
        elif self._fail_on_enter:
            self.set_state(GoalState.FAILED)


# --- Construction ---


def test_creation_basic():
    inner = MockGoal(name="inner")
    rep = GoalRepeater(inner, times=3, name="rep")
    assert rep._goal is inner
    assert rep._repeat_times == 3
    assert rep._times == 0
    assert rep.name == "rep"


def test_creation_propagates_max_duration_when_child_none():
    inner = MockGoal(name="inner", max_duration=None)
    GoalRepeater(inner, times=3, name="rep", max_duration=5.0)
    assert inner._max_duration == 5.0


def test_creation_propagates_max_duration_when_child_larger():
    inner = MockGoal(name="inner", max_duration=10.0)
    GoalRepeater(inner, times=3, name="rep", max_duration=5.0)
    assert inner._max_duration == 5.0


def test_creation_keeps_child_max_duration_when_smaller():
    inner = MockGoal(name="inner", max_duration=2.0)
    GoalRepeater(inner, times=3, name="rep", max_duration=5.0)
    assert inner._max_duration == 2.0


def test_creation_no_propagation_when_parent_max_none():
    inner = MockGoal(name="inner", max_duration=None)
    GoalRepeater(inner, times=3, name="rep", max_duration=None)
    assert inner._max_duration is None


# --- set_tick_freq ---


def test_set_tick_freq_propagates():
    inner = MockGoal(name="inner")
    rep = GoalRepeater(inner, times=2, name="rep")
    rep.set_tick_freq(100)
    assert inner._freq == 100


# --- serialize ---


def test_serialize():
    inner = MockGoal(name="inner")
    rep = GoalRepeater(inner, times=5, name="rep")
    data = rep.serialize()
    assert data["times"] == 5
    assert data["name"] == "rep"
    assert len(data["goals"]) == 1
    assert data["goals"][0]["name"] == "inner"


# --- on_enter (no crash) ---


def test_on_enter_no_crash():
    inner = MockGoal(name="inner")
    rep = GoalRepeater(inner, times=1, name="rep")
    rep.on_enter()


# --- enter: all iterations complete ---


def test_repeat_all_complete():
    inner = MockGoal(complete_on_enter=True, name="inner")
    rep = GoalRepeater(inner, times=3, name="rep")
    rep.enter()
    assert rep.state == GoalState.COMPLETED
    assert rep._times == 3


def test_repeat_single_iteration():
    inner = MockGoal(complete_on_enter=True, name="inner")
    rep = GoalRepeater(inner, times=1, name="rep")
    rep.enter()
    assert rep.state == GoalState.COMPLETED
    assert rep._times == 1


# --- enter: inner goal fails ---


def test_repeat_inner_fails():
    inner = MockGoal(
        complete_on_enter=False, fail_on_enter=True, name="inner", max_duration=0.01
    )
    rep = GoalRepeater(inner, times=3, name="rep", max_duration=1.0)
    rep.enter()
    assert rep.state == GoalState.FAILED


# --- enter: max_duration exceeded ---


def test_repeat_max_duration_exceeded():
    inner = MockGoal(complete_on_enter=True, name="inner")
    rep = GoalRepeater(inner, times=100, name="rep", max_duration=0.001)
    with patch.object(rep, "get_current_elapsed", return_value=1.0):
        rep.enter()
    assert rep.state == GoalState.FAILED


# --- enter: min_duration not met ---


def test_repeat_min_duration_not_met():
    inner = MockGoal(complete_on_enter=True, name="inner")
    rep = GoalRepeater(inner, times=1, name="rep", min_duration=1000.0)
    rep.enter()
    assert rep.state == GoalState.FAILED


# --- enter: terminated mid-loop ---


def test_repeat_terminated_mid_loop():
    call_count = 0

    class TerminatingGoal(Goal):
        def on_enter(self):
            pass

        def tick(self):
            nonlocal call_count
            call_count += 1
            self.set_state(GoalState.COMPLETED)

    inner = TerminatingGoal(name="inner")
    rep = GoalRepeater(inner, times=5, name="rep")

    original_enter = inner.enter

    def enter_and_terminate(*args, **kwargs):
        result = original_enter(*args, **kwargs)
        if rep._times >= 2:
            rep.set_state(GoalState.TERMINATED)
        return result

    with patch.object(inner, "enter", side_effect=enter_and_terminate):
        rep.enter()

    assert rep.state == GoalState.TERMINATED


# --- enter: returns self ---


def test_enter_returns_self():
    inner = MockGoal(complete_on_enter=True, name="inner")
    rep = GoalRepeater(inner, times=1, name="rep")
    result = rep.enter()
    assert result is rep


# --- enter: goal resets between iterations ---


def test_goal_resets_between_iterations():
    reset_count = 0

    class TrackingGoal(MockGoal):
        def reset(self):
            nonlocal reset_count
            reset_count += 1
            super().reset()

    inner = TrackingGoal(complete_on_enter=True, name="inner")
    rep = GoalRepeater(inner, times=3, name="rep")
    rep.enter()
    assert reset_count == 3


# --- enter: state is FAILED when inner fails mid-loop ---


def test_repeat_inner_fails_on_second_iteration():
    iteration = 0

    class FailOnSecond(Goal):
        def on_enter(self):
            pass

        def tick(self):
            nonlocal iteration
            iteration += 1
            if iteration >= 2:
                self.set_state(GoalState.FAILED)
            else:
                self.set_state(GoalState.COMPLETED)

    inner = FailOnSecond(name="inner", max_duration=0.1)
    rep = GoalRepeater(inner, times=3, name="rep", max_duration=1.0)
    rep.enter()
    assert rep.state == GoalState.FAILED


# --- enter: state breaks when already FAILED/COMPLETED ---


def test_repeat_breaks_when_state_not_running():
    inner = MockGoal(complete_on_enter=True, name="inner")
    rep = GoalRepeater(inner, times=5, name="rep")

    original_enter = inner.enter

    def enter_and_fail(*args, **kwargs):
        result = original_enter(*args, **kwargs)
        if rep._times >= 2:
            rep.set_state(GoalState.FAILED)
        return result

    with patch.object(inner, "enter", side_effect=enter_and_fail):
        rep.enter()

    assert rep.state == GoalState.FAILED
    assert rep._times < 5



"""Top-level package for goalee."""

from __future__ import annotations

__author__ = """Konstantinos Panayiotou"""
__email__ = "klpanagi@gmail.com"
__version__ = "0.2.1"

from goalee.brokers import AMQPBroker as AMQPBroker
from goalee.brokers import Broker as Broker
from goalee.brokers import MQTTBroker as MQTTBroker
from goalee.brokers import RedisBroker as RedisBroker
from goalee.entity import Entity as Entity
from goalee.scenario import Scenario as Scenario
from goalee.timing_goals import DeadlineGoal as DeadlineGoal
from goalee.timing_goals import LatencyGoal as LatencyGoal
from goalee.timing_goals import OrderingGoal as OrderingGoal
from goalee.timing_goals import RateGoal as RateGoal



from __future__ import annotations

import logging

from goalee import definitions as CONFIG

# Set up logging
logging.basicConfig(format="%(asctime)s - %(message)s")
if CONFIG.ZERO_LOGS:
    logging.disable()
else:
    logging.getLogger().setLevel(CONFIG.LOG_LEVEL)
# -------------------------------------------------------------

default_logger = logging.getLogger()



from __future__ import annotations

import math
import statistics
from enum import IntEnum
from typing import Any, Callable

from goalee.entity import Entity
from goalee.goal import Goal, GoalState

CONDITION_FUNCTIONS = {
    "std": statistics.stdev,
    "var": statistics.variance,
    "mean": statistics.mean,
    "min": min,
    "max": max,
    "fabs": math.fabs,
}


class EntityStateChange(Goal):
    def __init__(
        self,
        entity: Entity,
        name: str | None = None,
        event_emitter: Any | None = None,
        max_duration: float | None = None,
        min_duration: float | None = None,
        for_duration: float | None = None,
    ):
        super().__init__(
            [entity],
            event_emitter,
            name=name,
            max_duration=max_duration,
            min_duration=min_duration,
            for_duration=for_duration,
        )
        self.entity = entity
        self._last_state = self.entity.attributes.copy()

    def on_enter(self):
        self.log_debug(
            f"Starting EntityStateChange Goal <{self.name}>:\n"
            f"Entity: {self.entity.name}\n"
            f"Max Duration: {self._max_duration}\n"
            f"Min Duration: {self._min_duration}\n"
            f"For Duration: {self._for_duration}"
        )

    def tick(self):
        if self._last_state != self.entity.attributes:
            if self._for_duration is not None and self._for_duration > 0:
                self._ts_hold = self.get_current_ts()
            else:
                self.set_state(GoalState.COMPLETED)
        elif self._ts_hold is not None and self._for_duration is not None:
            if self.get_current_ts() - self._ts_hold > self._for_duration:
                self.set_state(GoalState.COMPLETED)
        self._last_state = self.entity.attributes.copy()


class EntityStateCondition(Goal):
    def __init__(
        self,
        entities: list[Entity],
        name: str | None = None,
        event_emitter: Any | None = None,
        condition: Callable | None = None,
        max_duration: float | None = None,
        min_duration: float | None = None,
        for_duration: float | None = None,
    ):
        super().__init__(
            entities,
            event_emitter,
            name=name,
            max_duration=max_duration,
            min_duration=min_duration,
            for_duration=for_duration,
        )
        self._condition = condition

    def get_entities_map(self):
        return {e.name: e for e in self._entities}

    def on_enter(self):
        self.log_debug(
            f"Starting EntityStateCondition Goal <{self.name}>:\n"
            f"Parameters:\n"
            f"  Condition: {self._condition}\n"
            f"  Max Duration: {self._max_duration}\n"
            f"  Min Duration: {self._min_duration}\n"
            f"  For Duration: {self._for_duration}"
        )
        self._ts_hold = -1.0

    def tick(self):
        """
        Evaluates the condition of the goal and updates its state accordingly.
        If the condition evaluates to True, the goal state is set to COMPLETED.

        Exceptions:
            - Catches and ignores TypeError exceptions.

        Raises:
            TypeError: If there is an issue with the type of the condition.
        """
        try:
            if isinstance(self._condition, type(lambda: None)):
                e = self.get_entities_map()
                cond_state = self._condition(e)
            elif callable(self._condition):
                cond_state = self._condition(self.get_entities_map())
            elif isinstance(self._condition, str):
                cond_state = self.evaluate_condition(self.get_entities_map())
            if cond_state:
                if self._for_duration is not None and self._for_duration > 0:
                    if self._ts_hold is None or self._ts_hold < 0:
                        self._ts_hold = self.get_current_ts()
                    elif self.get_current_ts() - self._ts_hold > self._for_duration:
                        self.set_state(GoalState.COMPLETED)
                else:
                    self.set_state(GoalState.COMPLETED)
            else:
                self._ts_hold = -1.0
        except TypeError:
            self.log_debug("Condition evaluation skipped (entity not yet initialized)")

    def evaluate_condition(self, entities):
        """
        Evaluates a condition based on the provided entities.

        This method uses the `eval` function to evaluate a condition stored in the
        instance variable `_condition`. The condition can use statistical functions
        such as standard deviation, variance, mean, min, and max, which are provided
        in the local scope for the evaluation.

        Args:
            entities (list): A list of entities to be used in the condition evaluation.

        Returns:
            bool: True if the condition evaluates to True, False otherwise. If an
                  exception occurs during evaluation, the method returns False.
        """
        try:
            # SAFETY: This eval() is intentional — it executes condition expressions
            # generated by the GoalDSL code generator. Do NOT use with untrusted input.
            return bool(
                eval(self._condition, {"entities": entities}, CONDITION_FUNCTIONS)
            )
        except Exception as e:
            if "NoneType" in str(e):
                pass
            else:
                self.log_error(f"Error in condition evaluation: {e}")
            return False


class AttrStreamStrategy(IntEnum):
    ALL = 0
    NONE = 1
    AT_LEAST_ONE = 2
    JUST_ONE = 3
    EXACTLY_X = 4
    ALL_ORDERED = 5
    EXACTLY_X_ORDERED = 6


class EntityAttrStream(Goal):
    def __init__(
        self,
        entity: list[Entity],
        attr: str,
        value: list[Any],
        strategy: AttrStreamStrategy,
        name: str | None = None,
        event_emitter: Any | None = None,
        max_duration: float | None = None,
        min_duration: float | None = None,
        for_duration: float | None = None,
    ):
        super().__init__(
            [entity],
            event_emitter,
            name=name,
            max_duration=max_duration,
            min_duration=min_duration,
            for_duration=for_duration,
        )
        self._entity = entity
        self._attr = attr
        self._value = value
        self._strategy = strategy
        self._last_state = None

        self._value_check_list = [False] * len(self._value)

    def on_enter(self):
        self.log_debug(
            f"Starting EntityAttrStream Goal <{self.name}>:\n"
            f"Attribute: {self._attr}\n"
            f"Value: {self._value}\n"
            f"Strategy: {self._strategy.name}\n"
            f"Max Duration: {self._max_duration}\n"
            f"Min Duration: {self._min_duration}"
        )

    def tick(self):
        _state = self._entity.attributes.copy()
        if self._last_state is None:
            self._last_state = _state
        elif _state[self._attr] == self._last_state[self._attr]:
            return

        self._last_state = _state

        if self._attr not in self._entity.attributes:
            raise ValueError(
                f"Attribute {self._attr} not found in entity {self._entity.name}"
            )

        for i, v in enumerate(self._value):
            if self._value_check_list[i]:
                continue
            if self._last_state[self._attr] == v:
                self._value_check_list[i] = True
                if self._strategy in (
                    AttrStreamStrategy.ALL_ORDERED,
                    AttrStreamStrategy.EXACTLY_X_ORDERED,
                ):
                    self.process_for_ordered_strategy()
                    break

        if self.is_done():
            self.set_state(GoalState.COMPLETED)

    def is_done(self):
        if self._strategy == AttrStreamStrategy.ALL:
            return all(self._value_check_list)
        elif self._strategy == AttrStreamStrategy.NONE:
            return not any(self._value_check_list)
        elif self._strategy == AttrStreamStrategy.AT_LEAST_ONE:
            return any(self._value_check_list)
        elif self._strategy == AttrStreamStrategy.JUST_ONE:
            return sum(self._value_check_list) == 1
        elif self._strategy == AttrStreamStrategy.EXACTLY_X:
            return sum(self._value_check_list) == len(self._value)
        elif self._strategy == AttrStreamStrategy.ALL_ORDERED:
            return all(self._value_check_list)
        elif self._strategy == AttrStreamStrategy.EXACTLY_X_ORDERED:
            return sum(self._value_check_list) == len(self._value)

    def process_for_ordered_strategy(self):
        idx = next(
            (
                i
                for i in reversed(range(len(self._value_check_list)))
                if self._value_check_list[i]
            ),
            -1,
        )
        if not all(self._value_check_list[:idx]):
            self.reset_check_list()

    # def process_for_ordered_strategy(self):
    #     self.log_info(f"Value check list: {self._value_check_list}")
    #     for i in range(len(self._value_check_list)-1):
    #         if self._value_check_list[i+1] and not self._value_check_list[i]:
    #             # self.reset_check_list()
    #             break

    def reset_check_list(self):
        self._value_check_list = [False] * len(self._value)



from __future__ import annotations

import math

from goalee.area_goals import (
    AreaGoalTag,
    CircularAreaGoal,
    MovingAreaGoal,
    RectangleAreaGoal,
)
from goalee.brokers import RedisBroker
from goalee.entity import Entity
from goalee.goal import GoalState
from goalee.types import Point


def make_entity(name="test_entity", attrs=None, source=None):
    attrs = attrs or ["position", "orientation"]
    source = source or RedisBroker()
    return Entity(
        name=name,
        etype="sensor",
        topic=f"{name}.topic",
        attributes=attrs,
        source=source,
    )


# ---------------------------------------------------------------------------
# AreaGoalTag enum
# ---------------------------------------------------------------------------


class TestAreaGoalTag:
    def test_values(self):
        assert AreaGoalTag.ENTER == 0
        assert AreaGoalTag.EXIT == 1
        assert AreaGoalTag.AVOID == 2
        assert AreaGoalTag.STEP == 3
        assert AreaGoalTag.STAY == 4
        assert AreaGoalTag.CROSS == 5


# ---------------------------------------------------------------------------
# RectangleAreaGoal
# ---------------------------------------------------------------------------


class TestRectangleAreaGoal:
    def _make_goal(self, entities=None, tag=AreaGoalTag.ENTER, for_duration=None):
        if entities is None:
            entities = [make_entity()]
        return RectangleAreaGoal(
            entities=entities,
            bottom_left_edge=Point(0.0, 0.0, 0.0),
            length_x=10.0,
            length_y=10.0,
            tag=tag,
            name="rect1",
            for_duration=for_duration,
        )

    def test_creation(self):
        goal = self._make_goal()
        assert goal.name == "rect1"
        assert goal.state == GoalState.IDLE
        assert goal._bottom_left_edge == Point(0.0, 0.0, 0.0)
        assert goal._length_x == 10.0
        assert goal._length_y == 10.0

    def test_tag_property(self):
        goal = self._make_goal(tag=AreaGoalTag.AVOID)
        assert goal.tag == AreaGoalTag.AVOID

    def test_on_enter_no_crash(self):
        goal = self._make_goal()
        goal.on_enter()

    def test_tick_updates_last_states(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": 5.0}
        goal = self._make_goal(entities=[entity])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal._last_states[0]["position"] == {"x": 5.0, "y": 5.0}

    def test_check_area_inside_enter_completes(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": 5.0}
        goal = self._make_goal(entities=[entity])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.COMPLETED

    def test_check_area_outside_enter_stays_running(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 15.0, "y": 15.0}
        goal = self._make_goal(entities=[entity])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_area_inside_avoid_fails(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": 5.0}
        goal = self._make_goal(entities=[entity], tag=AreaGoalTag.AVOID)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.FAILED

    def test_check_area_position_none_skips(self):
        entity = make_entity()
        goal = self._make_goal(entities=[entity])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_area_pos_x_none_skips(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": None, "y": 5.0}
        goal = self._make_goal(entities=[entity])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_area_pos_y_none_skips(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": None}
        goal = self._make_goal(entities=[entity])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_area_enter_with_for_duration_sets_ts_hold(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": 5.0}
        goal = self._make_goal(entities=[entity], for_duration=5.0)
        goal.set_state(GoalState.RUNNING)
        goal.get_current_ts = lambda: 100.0
        goal.tick()
        assert goal._ts_hold == 100.0
        assert goal.state == GoalState.RUNNING

    def test_check_area_enter_for_duration_completes(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": 5.0}
        goal = self._make_goal(entities=[entity], for_duration=5.0)
        goal.set_state(GoalState.RUNNING)
        goal.get_current_ts = lambda: 100.0
        goal.tick()
        goal.get_current_ts = lambda: 106.0
        goal.tick()
        assert goal.state == GoalState.COMPLETED

    def test_check_area_avoid_with_for_duration_sets_ts_hold(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": 5.0}
        goal = self._make_goal(
            entities=[entity], tag=AreaGoalTag.AVOID, for_duration=5.0
        )
        goal.set_state(GoalState.RUNNING)
        goal.get_current_ts = lambda: 100.0
        goal.tick()
        assert goal._ts_hold == 100.0
        assert goal.state == GoalState.RUNNING

    def test_check_area_avoid_for_duration_fails(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": 5.0}
        goal = self._make_goal(
            entities=[entity], tag=AreaGoalTag.AVOID, for_duration=5.0
        )
        goal.set_state(GoalState.RUNNING)
        goal.get_current_ts = lambda: 100.0
        goal.tick()
        goal.get_current_ts = lambda: 106.0
        goal.tick()
        assert goal.state == GoalState.FAILED

    def test_check_area_leaves_area_resets_ts_hold(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": 5.0}
        goal = self._make_goal(entities=[entity], for_duration=5.0)
        goal.set_state(GoalState.RUNNING)
        goal.get_current_ts = lambda: 100.0
        goal.tick()
        assert goal._ts_hold == 100.0
        entity.attributes["position"] = {"x": 15.0, "y": 15.0}
        goal.tick()
        assert goal._ts_hold == -1.0

    def test_check_area_on_boundary_stays_running(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 0.0, "y": 5.0}
        goal = self._make_goal(entities=[entity])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING


# ---------------------------------------------------------------------------
# CircularAreaGoal
# ---------------------------------------------------------------------------


class TestCircularAreaGoal:
    def _make_goal(self, entities=None, tag=AreaGoalTag.ENTER, for_duration=None):
        if entities is None:
            entities = [make_entity()]
        return CircularAreaGoal(
            entities=entities,
            center=Point(5.0, 5.0, 0.0),
            radius=3.0,
            tag=tag,
            name="circ1",
            for_duration=for_duration,
        )

    def test_creation(self):
        goal = self._make_goal()
        assert goal.name == "circ1"
        assert goal._center == Point(5.0, 5.0, 0.0)
        assert goal._radius == 3.0

    def test_tag_property(self):
        goal = self._make_goal(tag=AreaGoalTag.EXIT)
        assert goal.tag == AreaGoalTag.EXIT

    def test_on_enter_no_crash(self):
        goal = self._make_goal()
        goal.on_enter()

    def test_check_area_inside_enter_completes(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": 5.0}
        goal = self._make_goal(entities=[entity])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.COMPLETED

    def test_check_area_outside_enter_stays_running(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 50.0, "y": 50.0}
        goal = self._make_goal(entities=[entity])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_area_inside_avoid_fails(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": 5.0}
        goal = self._make_goal(entities=[entity], tag=AreaGoalTag.AVOID)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.FAILED

    def test_check_area_none_position_skips(self):
        entity = make_entity()
        goal = self._make_goal(entities=[entity])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_area_pos_x_none_skips(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": None, "y": 5.0}
        goal = self._make_goal(entities=[entity])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_area_enter_with_for_duration(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": 5.0}
        goal = self._make_goal(entities=[entity], for_duration=5.0)
        goal.set_state(GoalState.RUNNING)
        goal.get_current_ts = lambda: 100.0
        goal.tick()
        assert goal._ts_hold == 100.0
        assert goal.state == GoalState.RUNNING
        goal.get_current_ts = lambda: 106.0
        goal.tick()
        assert goal.state == GoalState.COMPLETED

    def test_check_area_avoid_with_for_duration(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": 5.0}
        goal = self._make_goal(
            entities=[entity], tag=AreaGoalTag.AVOID, for_duration=5.0
        )
        goal.set_state(GoalState.RUNNING)
        goal.get_current_ts = lambda: 100.0
        goal.tick()
        assert goal._ts_hold == 100.0
        goal.get_current_ts = lambda: 106.0
        goal.tick()
        assert goal.state == GoalState.FAILED

    def test_check_area_leaves_resets_ts_hold(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": 5.0}
        goal = self._make_goal(entities=[entity], for_duration=5.0)
        goal.set_state(GoalState.RUNNING)
        goal.get_current_ts = lambda: 100.0
        goal.tick()
        assert goal._ts_hold == 100.0
        entity.attributes["position"] = {"x": 50.0, "y": 50.0}
        goal.tick()
        assert goal._ts_hold == -1.0

    def test_calc_distance(self):
        goal = self._make_goal()
        pos = {"x": 8.0, "y": 9.0}
        expected = math.sqrt((8.0 - 5.0) ** 2 + (9.0 - 5.0) ** 2)
        assert goal._calc_distance(pos) == expected

    def test_tick_updates_states(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 1.0, "y": 1.0}
        goal = self._make_goal(entities=[entity])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal._last_states[0]["position"] == {"x": 1.0, "y": 1.0}

    def test_on_boundary_completes(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 8.0, "y": 5.0}
        goal = self._make_goal(entities=[entity])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.COMPLETED


# ---------------------------------------------------------------------------
# MovingAreaGoal
# ---------------------------------------------------------------------------


class TestMovingAreaGoal:
    def _make_goal(
        self,
        motion_entity=None,
        entities=None,
        tag=AreaGoalTag.ENTER,
        for_duration=None,
    ):
        if motion_entity is None:
            motion_entity = make_entity("motion")
        if entities is None:
            monitor = make_entity("monitor")
            entities = [monitor]
        return MovingAreaGoal(
            motion_entity=motion_entity,
            entities=entities,
            radius=3.0,
            tag=tag,
            name="moving1",
            for_duration=for_duration,
        )

    def test_creation(self):
        goal = self._make_goal()
        assert goal.name == "moving1"
        assert goal._radius == 3.0

    def test_motion_entity_removed_from_entities(self):
        motion = make_entity("motion")
        monitor = make_entity("monitor")
        entities = [motion, monitor]
        goal = MovingAreaGoal(
            motion_entity=motion,
            entities=entities,
            radius=3.0,
            name="moving2",
        )
        assert motion not in goal._entities
        assert monitor in goal._entities

    def test_motion_entity_not_in_entities_no_error(self):
        motion = make_entity("motion")
        monitor = make_entity("monitor")
        goal = MovingAreaGoal(
            motion_entity=motion,
            entities=[monitor],
            radius=3.0,
            name="moving3",
        )
        assert motion not in goal._entities

    def test_motion_entity_property(self):
        motion = make_entity("motion")
        goal = self._make_goal(motion_entity=motion)
        assert goal.motion_entity is motion

    def test_tag_property(self):
        goal = self._make_goal(tag=AreaGoalTag.AVOID)
        assert goal.tag == AreaGoalTag.AVOID

    def test_on_enter_no_crash(self):
        goal = self._make_goal()
        goal.on_enter()

    def test_check_area_motion_entity_state_none_returns(self):
        motion = make_entity("motion")
        monitor = make_entity("monitor")
        monitor.state = {"position": {"x": 5.0, "y": 5.0}}
        goal = self._make_goal(motion_entity=motion, entities=[monitor])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_area_motion_entity_state_empty_returns(self):
        motion = make_entity("motion")
        motion.state = {}
        monitor = make_entity("monitor")
        monitor.state = {"position": {"x": 5.0, "y": 5.0}}
        goal = self._make_goal(motion_entity=motion, entities=[monitor])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_area_monitor_state_none_skips(self):
        motion = make_entity("motion")
        motion.state = {"position": {"x": 5.0, "y": 5.0}}
        monitor = make_entity("monitor")
        goal = self._make_goal(motion_entity=motion, entities=[monitor])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_area_monitor_state_empty_skips(self):
        motion = make_entity("motion")
        motion.state = {"position": {"x": 5.0, "y": 5.0}}
        monitor = make_entity("monitor")
        monitor.state = {}
        goal = self._make_goal(motion_entity=motion, entities=[monitor])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_area_no_position_key_warns(self):
        motion = make_entity("motion")
        motion.state = {"position": {"x": 5.0, "y": 5.0}}
        monitor = make_entity("monitor")
        monitor.state = {"velocity": 1.0}
        goal = self._make_goal(motion_entity=motion, entities=[monitor])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_area_position_within_radius_enter_completes(self):
        motion = make_entity("motion")
        motion.state = {"position": {"x": 5.0, "y": 5.0}}
        monitor = make_entity("monitor")
        monitor.state = {"position": {"x": 6.0, "y": 6.0}}
        goal = self._make_goal(motion_entity=motion, entities=[monitor])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.COMPLETED

    def test_check_area_position_outside_radius_enter_stays_running(self):
        motion = make_entity("motion")
        motion.state = {"position": {"x": 5.0, "y": 5.0}}
        monitor = make_entity("monitor")
        monitor.state = {"position": {"x": 50.0, "y": 50.0}}
        goal = self._make_goal(motion_entity=motion, entities=[monitor])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_area_not_reached_avoid_fails(self):
        motion = make_entity("motion")
        motion.state = {"position": {"x": 5.0, "y": 5.0}}
        monitor = make_entity("monitor")
        monitor.state = {"position": {"x": 50.0, "y": 50.0}}
        goal = self._make_goal(
            motion_entity=motion, entities=[monitor], tag=AreaGoalTag.AVOID
        )
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.FAILED

    def test_check_area_enter_with_for_duration(self):
        motion = make_entity("motion")
        motion.state = {"position": {"x": 5.0, "y": 5.0}}
        monitor = make_entity("monitor")
        monitor.state = {"position": {"x": 6.0, "y": 6.0}}
        goal = self._make_goal(
            motion_entity=motion, entities=[monitor], for_duration=5.0
        )
        goal.set_state(GoalState.RUNNING)
        goal.get_current_ts = lambda: 100.0
        goal.tick()
        assert goal._ts_hold == 100.0
        assert goal.state == GoalState.RUNNING
        goal.get_current_ts = lambda: 106.0
        goal.tick()
        assert goal.state == GoalState.COMPLETED

    def test_check_area_avoid_with_for_duration(self):
        motion = make_entity("motion")
        motion.state = {"position": {"x": 5.0, "y": 5.0}}
        monitor = make_entity("monitor")
        monitor.state = {"position": {"x": 50.0, "y": 50.0}}
        goal = self._make_goal(
            motion_entity=motion,
            entities=[monitor],
            tag=AreaGoalTag.AVOID,
            for_duration=5.0,
        )
        goal.set_state(GoalState.RUNNING)
        goal.get_current_ts = lambda: 100.0
        goal.tick()
        assert goal._ts_hold == 100.0
        goal.get_current_ts = lambda: 106.0
        goal.tick()
        assert goal.state == GoalState.FAILED

    def test_check_area_leaves_area_resets_ts_hold(self):
        motion = make_entity("motion")
        motion.state = {"position": {"x": 5.0, "y": 5.0}}
        monitor = make_entity("monitor")
        monitor.state = {"position": {"x": 6.0, "y": 6.0}}
        goal = self._make_goal(
            motion_entity=motion, entities=[monitor], for_duration=5.0
        )
        goal.set_state(GoalState.RUNNING)
        goal.get_current_ts = lambda: 100.0
        goal.tick()
        assert goal._ts_hold == 100.0
        monitor.state = {"position": {"x": 50.0, "y": 50.0}}
        goal.tick()
        assert goal._ts_hold == -1.0

    def test_calc_distance(self):
        motion = make_entity("motion")
        motion.state = {"position": {"x": 5.0, "y": 5.0}}
        goal = self._make_goal(motion_entity=motion)
        pos = {"x": 8.0, "y": 9.0}
        expected = math.sqrt((8.0 - 5.0) ** 2 + (9.0 - 5.0) ** 2)
        assert goal._calc_distance(pos) == expected

    def test_check_area_pos_xy_none_warns(self):
        motion = make_entity("motion")
        motion.state = {"position": {"x": 5.0, "y": 5.0}}
        monitor = make_entity("monitor")
        monitor.state = {"position": {"x": None, "y": None}}
        goal = self._make_goal(motion_entity=motion, entities=[monitor])
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_reached_avoid_resets_ts_hold(self):
        motion = make_entity("motion")
        motion.state = {"position": {"x": 5.0, "y": 5.0}}
        monitor = make_entity("monitor")
        monitor.state = {"position": {"x": 6.0, "y": 6.0}}
        goal = self._make_goal(
            motion_entity=motion, entities=[monitor], tag=AreaGoalTag.AVOID
        )
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal._ts_hold == -1.0



from __future__ import annotations

from collections import deque
from typing import Any

from commlib.node import Node

from goalee.brokers import AMQPBroker, MQTTBroker, RedisBroker
from goalee.logging import default_logger as logger


# A class representing an entity communicating via an MQTT broker on a specific topic
class Entity:
    def __init__(
        self,
        name: str,
        etype: str,
        topic: str,
        attributes: list[str],
        source=None,
        init_buffers: bool = False,
        buffer_length: int = 10,
        strict_mode: bool = False,
    ) -> None:
        # Entity name
        self.name = name
        self.camel_name = self.to_camel_case(name)
        self.etype = etype
        # MQTT topic for Entity
        self.topic = topic
        self._strict = strict_mode
        # Entity state
        self.state = None
        # Set Entity's MQTT Broker
        self.source = source
        # Entity's Attributes
        self.attributes = {key: None for key in attributes}
        self.attributes_buff = {attr: None for attr in self.attributes}
        self.buffer_length = buffer_length
        if init_buffers:
            for attr in self.attributes:
                self.init_attr_buffer(attr, self.buffer_length)
        self._initialized = False
        self._started = False

    @property
    def initialized(self):
        return self._initialized

    def __getitem__(self, key):
        # Allow dictionary-style access to attributes
        # return getattr(self, key)
        return self.attributes[key]

    def get_buffer(self, attr_name: str, size: int = None):
        buf = self.attributes_buff[attr_name]
        if size is not None:
            return list(buf)[-size:]
        return list(buf)

    def get_attr(self, attr_name: str) -> Any:
        return self.attributes[attr_name]

    def init_attr_buffer(self, attr_name, size):
        self.attributes_buff[attr_name] = deque(maxlen=size)
        # self.attributes_buff[attr_name].extend([0] * size)

    def to_camel_case(self, snake_str):
        return "".join(x.capitalize() for x in snake_str.lower().split("_"))

    def create_node(self):
        if self.source is None:
            raise ValueError(f"Entity {self.name} not assigned a broker")
        if isinstance(self.source, RedisBroker):
            from commlib.transports.redis import ConnectionParameters

            conn_params = ConnectionParameters(
                host=self.source.host,
                port=self.source.port,
                db=self.source.db,
                username=self.source.username,
                password=self.source.password,
            )
        elif isinstance(self.source, AMQPBroker):
            from commlib.transports.amqp import ConnectionParameters

            conn_params = ConnectionParameters(
                host=self.source.host,
                port=self.source.port,
                vhost=self.source.vhost,
                username=self.source.username,
                password=self.source.password,
            )
        elif isinstance(self.source, MQTTBroker):
            from commlib.transports.mqtt import ConnectionParameters

            conn_params = ConnectionParameters(
                host=self.source.host,
                port=self.source.port,
                username=self.source.username,
                password=self.source.password,
            )
        else:
            raise ValueError("Invalid broker type")
        self.conn_params = conn_params

        self.node = Node(
            node_name=self.camel_name,
            connection_params=self.conn_params,
            debug=False,
            heartbeats=False,
        )
        self.subscriber = self.node.create_subscriber(
            topic=self.topic, on_message=self.update_state
        )

    def start(self):
        """
        Starts the entity by creating a node, setting up a subscriber to listen to messages on the entity's topic,
        and running the subscriber. Additionally, it creates a publisher for communications on the same topic.

        The subscriber will call the `update_state` method whenever a message is received on the topic.
        """
        if self._started:
            return
        self._started = True
        self.create_node()
        self.node.run()
        logger.info(f"Started Entity <{self.name}> listening on topic <{self.topic}>")

    def update_state(self, new_state: dict[str, Any]) -> None:
        """
        Function for updating Entity state. Meant to be used as a callback function by the Entity's subscriber object
        (commlib-py).
        :param new_state: Dictionary containing the Entity's state
        :return:
        """
        # Update state
        # logger.info(f'[Entity {self.name}] State Change')
        state = new_state.copy()
        for key in state:
            if key not in self.attributes:
                # logger.warning(
                #     f"Entity <{self.name}> state - Message KeyError <{key}>\n"
                # )
                if self._strict:
                    logger.warning(
                        f"Entity <{self.name}> in strict mode - Dropping invalid message"
                    )
                    return
                else:
                    continue
        self._initialized = True
        self.state = state
        # Update attributes based on state
        self.update_attributes(state)
        self.update_buffers(state)

    def update_buffers(self, new_state):
        """
        Recursive function used by update_state() mainly to updated
            dictionaries/objects and normal Attributes.
        """
        # Update attributes
        state = new_state.copy()
        for key in state:
            if self._strict and key not in self.attributes:
                logger.warning(
                    f"Entity <{self.name}> in strict mode - Dropping invalid message"
                )
                return
        for attribute, value in state.items():
            # If value is a dictionary, also update the Dict's subattributes/items
            if (
                attribute in self.attributes
                and self.attributes_buff[attribute] is not None
            ):
                self.attributes_buff[attribute].append(value)

    def update_attributes(self, new_state):
        """
        Recursive function used by update_state() mainly to updated
            dictionaries/objects and normal Attributes.
        """
        # Update attributes
        for key, value in new_state.items():
            # If value is a dictionary, also update the Dict's subattributes/items
            # if root[attribute].__class__.__name__ == 'TimeAttribute':
            #     setattr(root[attribute].value, 'hour', value['hour'])
            #     setattr(root[attribute].value, 'minute', value['minute'])
            #     setattr(root[attribute].value, 'second', value['second'])
            if key in self.attributes:
                self.attributes[key] = value



from __future__ import annotations

import pytest

from goalee.types import Orientation, Point, Pose

# --- Point ---


class TestPointSub:
    def test_sub_point(self):
        p1 = Point(5.0, 7.0, 9.0)
        p2 = Point(1.0, 2.0, 3.0)
        result = p1 - p2
        assert result == Point(4.0, 5.0, 6.0)

    def test_sub_float(self):
        p = Point(5.0, 7.0, 9.0)
        result = p - 1.0
        assert result == Point(4.0, 6.0, 8.0)

    def test_sub_int(self):
        p = Point(5.0, 7.0, 9.0)
        result = p - 1
        assert result == Point(4.0, 6.0, 8.0)

    def test_sub_invalid_raises(self):
        p = Point(1.0, 2.0, 3.0)
        with pytest.raises(ValueError):
            p - "invalid"


class TestPointEq:
    def test_eq_point_true(self):
        assert Point(1.0, 2.0, 3.0) == Point(1.0, 2.0, 3.0)

    def test_eq_point_false(self):
        assert Point(1.0, 2.0, 3.0) != Point(4.0, 5.0, 6.0)

    def test_eq_dict_true(self):
        p = Point(1.0, 2.0, 3.0)
        assert p == {"x": 1.0, "y": 2.0, "z": 3.0}

    def test_eq_dict_false(self):
        p = Point(1.0, 2.0, 3.0)
        assert p != {"x": 9.0, "y": 9.0, "z": 9.0}

    def test_eq_other_returns_false(self):
        p = Point(1.0, 2.0, 3.0)
        assert (p == "string") is False
        assert (p == 42) is False


class TestPointComparisons:
    def test_lt_point(self):
        assert Point(1.0, 2.0, 3.0) < Point(2.0, 2.0, 3.0)
        assert not (Point(2.0, 2.0, 3.0) < Point(1.0, 2.0, 3.0))

    def test_lt_dict(self):
        assert Point(1.0, 2.0, 3.0) < {"x": 2.0, "y": 2.0, "z": 3.0}

    def test_lt_unsupported(self):
        assert Point(1.0, 2.0, 3.0).__lt__("bad") is NotImplemented

    def test_le_point(self):
        assert Point(1.0, 2.0, 3.0) <= Point(1.0, 2.0, 3.0)
        assert Point(1.0, 2.0, 3.0) <= Point(2.0, 2.0, 3.0)

    def test_le_dict(self):
        assert Point(1.0, 2.0, 3.0) <= {"x": 1.0, "y": 2.0, "z": 3.0}

    def test_le_unsupported(self):
        assert Point(1.0, 2.0, 3.0).__le__("bad") is NotImplemented

    def test_gt_point(self):
        assert Point(2.0, 2.0, 3.0) > Point(1.0, 2.0, 3.0)
        assert not (Point(1.0, 2.0, 3.0) > Point(2.0, 2.0, 3.0))

    def test_gt_dict(self):
        assert Point(2.0, 2.0, 3.0) > {"x": 1.0, "y": 2.0, "z": 3.0}

    def test_gt_unsupported(self):
        assert Point(1.0, 2.0, 3.0).__gt__("bad") is NotImplemented

    def test_ge_point(self):
        assert Point(1.0, 2.0, 3.0) >= Point(1.0, 2.0, 3.0)
        assert Point(2.0, 2.0, 3.0) >= Point(1.0, 2.0, 3.0)

    def test_ge_dict(self):
        assert Point(1.0, 2.0, 3.0) >= {"x": 1.0, "y": 2.0, "z": 3.0}

    def test_ge_unsupported(self):
        assert Point(1.0, 2.0, 3.0).__ge__("bad") is NotImplemented


class TestPointAbs:
    def test_abs_known_values(self):
        p = Point(3.0, 4.0, 0.0)
        assert p.abs() == 5.0

    def test_abs_origin(self):
        p = Point(0.0, 0.0, 0.0)
        assert p.abs() == 0.0

    def test_abs_3d(self):
        p = Point(1.0, 2.0, 2.0)
        assert p.abs() == 3.0


class TestPointAdd:
    def test_add_point(self):
        result = Point(1.0, 2.0, 3.0) + Point(4.0, 5.0, 6.0)
        assert result == Point(5.0, 7.0, 9.0)

    def test_add_float(self):
        result = Point(1.0, 2.0, 3.0) + 10.0
        assert result == Point(11.0, 12.0, 13.0)

    def test_add_int(self):
        result = Point(1.0, 2.0, 3.0) + 10
        assert result == Point(11.0, 12.0, 13.0)

    def test_add_invalid_raises(self):
        with pytest.raises(ValueError):
            Point(1.0, 2.0, 3.0) + "invalid"


class TestPointDefaults:
    def test_defaults(self):
        p = Point()
        assert p.x == 0.0
        assert p.y == 0.0
        assert p.z == 0.0


# --- Orientation ---


class TestOrientationSub:
    def test_sub_orientation(self):
        o1 = Orientation(0.5, 0.7, 0.9)
        o2 = Orientation(0.1, 0.2, 0.3)
        result = o1 - o2
        assert abs(result.roll - 0.4) < 1e-9
        assert abs(result.pitch - 0.5) < 1e-9
        assert abs(result.yaw - 0.6) < 1e-9

    def test_sub_float(self):
        o = Orientation(1.0, 2.0, 3.0)
        result = o - 0.5
        assert abs(result.roll - 0.5) < 1e-9
        assert abs(result.pitch - 1.5) < 1e-9
        assert abs(result.yaw - 2.5) < 1e-9

    def test_sub_int(self):
        o = Orientation(1.0, 2.0, 3.0)
        result = o - 1
        assert abs(result.roll - 0.0) < 1e-9

    def test_sub_invalid_raises(self):
        with pytest.raises(ValueError):
            Orientation(1.0, 2.0, 3.0) - "invalid"


class TestOrientationEq:
    def test_eq_orientation_true(self):
        assert Orientation(0.1, 0.2, 0.3) == Orientation(0.1, 0.2, 0.3)

    def test_eq_orientation_false(self):
        assert Orientation(0.1, 0.2, 0.3) != Orientation(0.4, 0.5, 0.6)

    def test_eq_dict_true(self):
        o = Orientation(0.1, 0.2, 0.3)
        assert o == {"roll": 0.1, "pitch": 0.2, "yaw": 0.3}

    def test_eq_dict_false(self):
        o = Orientation(0.1, 0.2, 0.3)
        assert o != {"roll": 9.0, "pitch": 9.0, "yaw": 9.0}

    def test_eq_other_returns_false(self):
        o = Orientation(0.1, 0.2, 0.3)
        assert (o == "string") is False
        assert (o == 42) is False


class TestOrientationAbs:
    def test_abs_known(self):
        o = Orientation(1.0, 2.0, 2.0)
        assert o.abs() == 3.0

    def test_abs_zero(self):
        o = Orientation(0.0, 0.0, 0.0)
        assert o.abs() == 0.0


class TestOrientationAdd:
    def test_add_orientation(self):
        result = Orientation(0.1, 0.2, 0.3) + Orientation(0.4, 0.5, 0.6)
        assert abs(result.roll - 0.5) < 1e-9
        assert abs(result.pitch - 0.7) < 1e-9
        assert abs(result.yaw - 0.9) < 1e-9

    def test_add_float(self):
        result = Orientation(0.1, 0.2, 0.3) + 1.0
        assert abs(result.roll - 1.1) < 1e-9
        assert abs(result.pitch - 1.2) < 1e-9
        assert abs(result.yaw - 1.3) < 1e-9

    def test_add_int(self):
        result = Orientation(0.1, 0.2, 0.3) + 1
        assert abs(result.roll - 1.1) < 1e-9

    def test_add_invalid_raises(self):
        with pytest.raises(ValueError):
            Orientation(0.1, 0.2, 0.3) + "invalid"


class TestOrientationComparisons:
    def test_gt_orientation(self):
        assert Orientation(1.0, 0.0, 0.0) > Orientation(0.0, 0.0, 0.0)
        assert not (Orientation(0.0, 0.0, 0.0) > Orientation(1.0, 0.0, 0.0))

    def test_gt_dict(self):
        assert Orientation(1.0, 0.0, 0.0) > {"roll": 0.0, "pitch": 0.0, "yaw": 0.0}

    def test_gt_unsupported(self):
        assert Orientation(1.0, 0.0, 0.0).__gt__("bad") is NotImplemented

    def test_ge_orientation(self):
        assert Orientation(1.0, 0.0, 0.0) >= Orientation(1.0, 0.0, 0.0)
        assert Orientation(1.0, 0.0, 0.0) >= Orientation(0.0, 0.0, 0.0)

    def test_ge_dict(self):
        assert Orientation(1.0, 0.0, 0.0) >= {"roll": 1.0, "pitch": 0.0, "yaw": 0.0}

    def test_ge_unsupported(self):
        assert Orientation(1.0, 0.0, 0.0).__ge__("bad") is NotImplemented

    def test_lt_orientation(self):
        assert Orientation(0.0, 0.0, 0.0) < Orientation(1.0, 0.0, 0.0)
        assert not (Orientation(1.0, 0.0, 0.0) < Orientation(0.0, 0.0, 0.0))

    def test_lt_dict(self):
        assert Orientation(0.0, 0.0, 0.0) < {"roll": 1.0, "pitch": 0.0, "yaw": 0.0}

    def test_lt_unsupported(self):
        assert Orientation(0.0, 0.0, 0.0).__lt__("bad") is NotImplemented

    def test_le_orientation(self):
        assert Orientation(0.0, 0.0, 0.0) <= Orientation(0.0, 0.0, 0.0)
        assert Orientation(0.0, 0.0, 0.0) <= Orientation(1.0, 0.0, 0.0)

    def test_le_dict(self):
        assert Orientation(0.0, 0.0, 0.0) <= {"roll": 0.0, "pitch": 0.0, "yaw": 0.0}

    def test_le_unsupported(self):
        # Orientation.__le__ with dict is the last branch — no return NotImplemented
        # The source code has no return for unsupported in __le__, so it returns None
        result = Orientation(0.0, 0.0, 0.0).__le__("bad")
        assert result is None


class TestOrientationXYZAliases:
    def test_xyz_alias_single(self):
        o = Orientation(z=1.5)
        assert o.yaw == 1.5
        assert o.roll == 0.0
        assert o.pitch == 0.0

    def test_xyz_alias_all(self):
        o = Orientation(x=0.1, y=0.2, z=0.3)
        assert o.roll == 0.1
        assert o.pitch == 0.2
        assert o.yaw == 0.3

    def test_roll_pitch_yaw_still_work(self):
        o = Orientation(roll=0.1, pitch=0.2, yaw=0.3)
        assert o.roll == 0.1
        assert o.pitch == 0.2
        assert o.yaw == 0.3

    def test_explicit_roll_pitch_yaw_takes_precedence(self):
        o = Orientation(roll=0.5, x=0.9)
        assert o.roll == 0.5

    def test_positional_args_still_work(self):
        o = Orientation(0.1, 0.2, 0.3)
        assert o.roll == 0.1
        assert o.pitch == 0.2
        assert o.yaw == 0.3


class TestOrientationDefaults:
    def test_defaults(self):
        o = Orientation()
        assert o.roll == 0.0
        assert o.pitch == 0.0
        assert o.yaw == 0.0


# --- Pose ---


class TestPose:
    def test_creation(self):
        p = Point(1.0, 2.0, 3.0)
        o = Orientation(0.1, 0.2, 0.3)
        pose = Pose(translation=p, orientation=o)
        assert pose.translation == p
        assert pose.orientation == o

    def test_pose_fields(self):
        pose = Pose(translation=Point(), orientation=Orientation())
        assert pose.translation.x == 0.0
        assert pose.orientation.roll == 0.0



from __future__ import annotations

from goalee.brokers import RedisBroker
from goalee.entity import Entity
from goalee.goal import GoalState
from goalee.pose_goals import OrientationGoal, PoseGoal, PositionGoal
from goalee.types import Orientation, Point


def make_entity(name="test_entity", attrs=None, source=None):
    attrs = attrs or ["position", "orientation"]
    source = source or RedisBroker()
    return Entity(
        name=name,
        etype="sensor",
        topic=f"{name}.topic",
        attributes=attrs,
        source=source,
    )


# ---------------------------------------------------------------------------
# PoseGoal
# ---------------------------------------------------------------------------


class TestPoseGoal:
    def _make_goal(self, entity=None, deviation_pos=1.0, deviation_ori=0.5):
        if entity is None:
            entity = make_entity()
        return PoseGoal(
            entity=entity,
            position=Point(5.0, 5.0, 0.0),
            orientation=Orientation(0.0, 0.0, 0.0),
            deviation_pos=deviation_pos,
            deviation_ori=deviation_ori,
            name="pose1",
        )

    def test_creation(self):
        goal = self._make_goal()
        assert goal.name == "pose1"
        assert goal._position == Point(5.0, 5.0, 0.0)
        assert goal._orientation == Orientation(0.0, 0.0, 0.0)
        assert goal._deviation_pos == 1.0
        assert goal._deviation_ori == 0.5

    def test_on_enter_no_crash(self):
        goal = self._make_goal()
        goal.on_enter()

    def test_tick_updates_last_state(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": 5.0, "z": 0.0}
        entity.attributes["orientation"] = {
            "x": 0.0,
            "y": 0.0,
            "roll": 0.0,
            "pitch": 0.0,
            "yaw": 0.0,
        }
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal._last_state["position"] == {"x": 5.0, "y": 5.0, "z": 0.0}

    def test_check_pose_no_position_returns(self):
        entity = make_entity()
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_pose_position_x_none_returns(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": None, "y": 5.0, "z": 0.0}
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_pose_position_y_none_returns(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": None, "z": 0.0}
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_pose_no_orientation_returns(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": 5.0, "z": 0.0}
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_pose_orientation_x_none_returns(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": 5.0, "z": 0.0}
        entity.attributes["orientation"] = {"x": None, "y": 0.0}
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_pose_orientation_y_none_returns(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.0, "y": 5.0, "z": 0.0}
        entity.attributes["orientation"] = {"x": 0.0, "y": None}
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_pose_within_deviation_completes(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.5, "y": 5.5, "z": 0.5}
        entity.attributes["orientation"] = {
            "x": 0.1,
            "y": 0.1,
            "roll": 0.1,
            "pitch": 0.1,
            "yaw": 0.1,
        }
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.COMPLETED

    def test_check_pose_outside_deviation_stays_running(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 50.0, "y": 50.0, "z": 50.0}
        entity.attributes["orientation"] = {
            "x": 0.0,
            "y": 0.0,
            "roll": 0.0,
            "pitch": 0.0,
            "yaw": 0.0,
        }
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING


# ---------------------------------------------------------------------------
# PositionGoal
# ---------------------------------------------------------------------------


class TestPositionGoal:
    def _make_goal(self, entity=None, deviation=1.0):
        if entity is None:
            entity = make_entity()
        return PositionGoal(
            entity=entity,
            position=Point(5.0, 5.0, 0.0),
            deviation=deviation,
            name="pos1",
        )

    def test_creation(self):
        goal = self._make_goal()
        assert goal.name == "pos1"
        assert goal._position == Point(5.0, 5.0, 0.0)
        assert goal._deviation == 1.0

    def test_on_enter_no_crash(self):
        goal = self._make_goal()
        goal.on_enter()

    def test_check_pos_no_position_returns(self):
        entity = make_entity()
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_pos_all_xyz_none_warns(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": None, "y": None, "z": None}
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_pos_within_deviation_completes(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 5.5, "y": 5.5, "z": 0.5}
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.COMPLETED

    def test_check_pos_outside_deviation_stays_running(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 50.0, "y": 50.0, "z": 50.0}
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_pos_on_boundary_stays_running(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 6.0, "y": 5.0, "z": 0.0}
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_tick_updates_last_state(self):
        entity = make_entity()
        entity.attributes["position"] = {"x": 1.0, "y": 2.0, "z": 3.0}
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal._last_state["position"] == {"x": 1.0, "y": 2.0, "z": 3.0}


# ---------------------------------------------------------------------------
# OrientationGoal
# ---------------------------------------------------------------------------


class TestOrientationGoal:
    def _make_goal(self, entity=None, deviation=0.5):
        if entity is None:
            entity = make_entity()
        return OrientationGoal(
            entity=entity,
            orientation=Orientation(0.0, 0.0, 0.0),
            deviation=deviation,
            name="ori1",
        )

    def test_creation(self):
        goal = self._make_goal()
        assert goal.name == "ori1"
        assert goal._orientation == Orientation(0.0, 0.0, 0.0)
        assert goal._deviation == 0.5

    def test_on_enter_no_crash(self):
        goal = self._make_goal()
        goal.on_enter()

    def test_check_ori_no_orientation_returns(self):
        entity = make_entity()
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_ori_all_none_returns(self):
        entity = make_entity()
        entity.attributes["orientation"] = {"roll": None, "pitch": None, "yaw": None}
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_check_ori_within_deviation_completes(self):
        entity = make_entity()
        entity.attributes["orientation"] = {"roll": 0.1, "pitch": 0.1, "yaw": 0.1}
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.COMPLETED

    def test_check_ori_outside_deviation_stays_running(self):
        entity = make_entity()
        entity.attributes["orientation"] = {"roll": 5.0, "pitch": 5.0, "yaw": 5.0}
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING

    def test_tick_updates_last_state(self):
        entity = make_entity()
        entity.attributes["orientation"] = {"roll": 0.1, "pitch": 0.2, "yaw": 0.3}
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal._last_state["orientation"] == {
            "roll": 0.1,
            "pitch": 0.2,
            "yaw": 0.3,
        }

    def test_check_ori_on_boundary_stays_running(self):
        entity = make_entity()
        entity.attributes["orientation"] = {"roll": 0.5, "pitch": 0.0, "yaw": 0.0}
        goal = self._make_goal(entity=entity)
        goal.set_state(GoalState.RUNNING)
        goal.tick()
        assert goal.state == GoalState.RUNNING



from __future__ import annotations

from collections import deque
from unittest.mock import MagicMock, patch

import pytest

from goalee.brokers import AMQPBroker, MQTTBroker, RedisBroker
from goalee.entity import Entity


class TestEntityCreation:
    def test_basic_creation(self):
        broker = RedisBroker()
        e = Entity(
            name="sensor1",
            etype="sensor",
            topic="sensors.temp",
            attributes=["temp", "humidity"],
            source=broker,
        )
        assert e.name == "sensor1"
        assert e.etype == "sensor"
        assert e.topic == "sensors.temp"
        assert e.source is broker
        assert e._strict is False
        assert e._initialized is False
        assert e._started is False

    def test_attributes_initialized_to_none(self):
        e = Entity(name="e", etype="t", topic="t", attributes=["a", "b", "c"])
        assert e.attributes == {"a": None, "b": None, "c": None}

    def test_init_buffers_true(self):
        e = Entity(
            name="e",
            etype="t",
            topic="t",
            attributes=["a", "b"],
            init_buffers=True,
            buffer_length=5,
        )
        assert isinstance(e.attributes_buff["a"], deque)
        assert e.attributes_buff["a"].maxlen == 5
        assert isinstance(e.attributes_buff["b"], deque)
        assert e.attributes_buff["b"].maxlen == 5

    def test_init_buffers_false(self):
        e = Entity(name="e", etype="t", topic="t", attributes=["a", "b"])
        assert e.attributes_buff["a"] is None
        assert e.attributes_buff["b"] is None

    def test_strict_mode(self):
        e = Entity(
            name="e",
            etype="t",
            topic="t",
            attributes=["a"],
            strict_mode=True,
        )
        assert e._strict is True


class TestEntityGetItem:
    def test_getitem(self):
        e = Entity(name="e", etype="t", topic="t", attributes=["temp"])
        e.attributes["temp"] = 25.0
        assert e["temp"] == 25.0

    def test_getitem_none(self):
        e = Entity(name="e", etype="t", topic="t", attributes=["temp"])
        assert e["temp"] is None


class TestEntityGetAttr:
    def test_get_attr(self):
        e = Entity(name="e", etype="t", topic="t", attributes=["temp"])
        e.attributes["temp"] = 42.0
        assert e.get_attr("temp") == 42.0


class TestEntityToCamelCase:
    def test_snake_to_camel(self):
        e = Entity(name="my_entity_name", etype="t", topic="t", attributes=[])
        assert e.camel_name == "MyEntityName"
        assert e.to_camel_case("my_entity") == "MyEntity"

    def test_single_word(self):
        e = Entity(name="sensor", etype="t", topic="t", attributes=[])
        assert e.camel_name == "Sensor"

    def test_already_camel(self):
        e = Entity(name="MySensor", etype="t", topic="t", attributes=[])
        assert e.camel_name == "Mysensor"


class TestEntityInitialized:
    def test_initially_false(self):
        e = Entity(name="e", etype="t", topic="t", attributes=["a"])
        assert e.initialized is False

    def test_true_after_update(self):
        e = Entity(name="e", etype="t", topic="t", attributes=["a"])
        e.update_state({"a": 1})
        assert e.initialized is True


class TestEntityUpdateState:
    def test_updates_attributes(self):
        e = Entity(name="e", etype="t", topic="t", attributes=["a", "b"])
        e.update_state({"a": 10, "b": 20})
        assert e.attributes["a"] == 10
        assert e.attributes["b"] == 20
        assert e._initialized is True

    def test_unknown_key_non_strict_continues(self):
        e = Entity(name="e", etype="t", topic="t", attributes=["a"])
        e.update_state({"a": 10, "unknown": 99})
        assert e.attributes["a"] == 10
        assert e._initialized is True

    def test_unknown_key_strict_drops(self):
        e = Entity(
            name="e",
            etype="t",
            topic="t",
            attributes=["a"],
            strict_mode=True,
        )
        e.update_state({"a": 10, "unknown": 99})
        assert e.attributes["a"] is None
        assert e._initialized is False


class TestEntityUpdateAttributes:
    def test_updates_known_keys(self):
        e = Entity(name="e", etype="t", topic="t", attributes=["x", "y"])
        e.update_attributes({"x": 1, "y": 2})
        assert e.attributes["x"] == 1
        assert e.attributes["y"] == 2

    def test_ignores_unknown_keys(self):
        e = Entity(name="e", etype="t", topic="t", attributes=["x"])
        e.update_attributes({"x": 1, "z": 99})
        assert e.attributes["x"] == 1
        assert "z" not in e.attributes


class TestEntityUpdateBuffers:
    def test_appends_to_buffer(self):
        e = Entity(
            name="e",
            etype="t",
            topic="t",
            attributes=["a"],
            init_buffers=True,
            buffer_length=3,
        )
        e.update_buffers({"a": 1})
        e.update_buffers({"a": 2})
        assert list(e.attributes_buff["a"]) == [1, 2]

    def test_no_append_when_buffer_none(self):
        e = Entity(name="e", etype="t", topic="t", attributes=["a"])
        e.update_buffers({"a": 1})
        assert e.attributes_buff["a"] is None

    def test_strict_mode_drops_invalid(self):
        e = Entity(
            name="e",
            etype="t",
            topic="t",
            attributes=["a"],
            init_buffers=True,
            strict_mode=True,
        )
        e.update_buffers({"a": 1, "bad": 2})
        assert len(e.attributes_buff["a"]) == 0


class TestEntityGetBuffer:
    def test_get_buffer_full(self):
        e = Entity(
            name="e",
            etype="t",
            topic="t",
            attributes=["a"],
            init_buffers=True,
            buffer_length=3,
        )
        e.attributes_buff["a"].append(1)
        e.attributes_buff["a"].append(2)
        e.attributes_buff["a"].append(3)
        result = e.get_buffer("a")
        assert result == [1, 2, 3]

    def test_get_buffer_with_size(self):
        e = Entity(
            name="e",
            etype="t",
            topic="t",
            attributes=["a"],
            init_buffers=True,
            buffer_length=5,
        )
        for i in range(5):
            e.attributes_buff["a"].append(i)
        result = e.get_buffer("a", size=3)
        assert result == [2, 3, 4]

    def test_get_buffer_not_full_returns_actual_values(self):
        e = Entity(
            name="e",
            etype="t",
            topic="t",
            attributes=["a"],
            init_buffers=True,
            buffer_length=5,
        )
        e.attributes_buff["a"].append(1)
        result = e.get_buffer("a")
        assert result == [1]

    def test_get_buffer_empty_returns_empty_list(self):
        e = Entity(
            name="e",
            etype="t",
            topic="t",
            attributes=["a"],
            init_buffers=True,
            buffer_length=5,
        )
        result = e.get_buffer("a")
        assert result == []

    def test_get_buffer_partial_fill_with_size(self):
        e = Entity(
            name="e",
            etype="t",
            topic="t",
            attributes=["a"],
            init_buffers=True,
            buffer_length=10,
        )
        for v in [1, 2, 3]:
            e.attributes_buff["a"].append(v)
        result = e.get_buffer("a", size=2)
        assert result == [2, 3]


class TestEntityInitAttrBuffer:
    def test_creates_deque(self):
        e = Entity(name="e", etype="t", topic="t", attributes=["a"])
        e.init_attr_buffer("a", 10)
        assert isinstance(e.attributes_buff["a"], deque)
        assert e.attributes_buff["a"].maxlen == 10


class TestEntityCreateNode:
    def test_no_source_raises(self):
        e = Entity(name="e", etype="t", topic="t", attributes=["a"], source=None)
        with pytest.raises(ValueError, match="not assigned a broker"):
            e.create_node()

    def test_invalid_broker_type_raises(self):
        e = Entity(name="e", etype="t", topic="t", attributes=["a"], source="invalid")
        with pytest.raises(ValueError, match="Invalid broker type"):
            e.create_node()

    @patch("goalee.entity.Node")
    def test_redis_broker(self, mock_node_cls):
        mock_node = MagicMock()
        mock_node_cls.return_value = mock_node
        broker = RedisBroker(host="redis-host", port=6380, db=1)
        e = Entity(name="e", etype="t", topic="t", attributes=["a"], source=broker)

        with patch.dict("sys.modules", {"commlib.transports.redis": MagicMock()}):
            e.create_node()

        assert e.node is mock_node
        mock_node.create_subscriber.assert_called_once()

    @patch("goalee.entity.Node")
    def test_amqp_broker(self, mock_node_cls):
        mock_node = MagicMock()
        mock_node_cls.return_value = mock_node
        broker = AMQPBroker(host="amqp-host", port=5673)
        e = Entity(name="e", etype="t", topic="t", attributes=["a"], source=broker)

        with patch.dict("sys.modules", {"commlib.transports.amqp": MagicMock()}):
            e.create_node()

        assert e.node is mock_node

    @patch("goalee.entity.Node")
    def test_mqtt_broker(self, mock_node_cls):
        mock_node = MagicMock()
        mock_node_cls.return_value = mock_node
        broker = MQTTBroker(host="mqtt-host", port=1884)
        e = Entity(name="e", etype="t", topic="t", attributes=["a"], source=broker)

        with patch.dict("sys.modules", {"commlib.transports.mqtt": MagicMock()}):
            e.create_node()

        assert e.node is mock_node


class TestEntityStart:
    @patch("goalee.entity.Node")
    def test_start_sets_started(self, mock_node_cls):
        mock_node = MagicMock()
        mock_node_cls.return_value = mock_node
        broker = RedisBroker()
        e = Entity(name="e", etype="t", topic="t", attributes=["a"], source=broker)

        with patch.dict("sys.modules", {"commlib.transports.redis": MagicMock()}):
            e.start()

        assert e._started is True
        mock_node.run.assert_called_once()

    @patch("goalee.entity.Node")
    def test_start_twice_is_noop(self, mock_node_cls):
        mock_node = MagicMock()
        mock_node_cls.return_value = mock_node
        broker = RedisBroker()
        e = Entity(name="e", etype="t", topic="t", attributes=["a"], source=broker)

        with patch.dict("sys.modules", {"commlib.transports.redis": MagicMock()}):
            e.start()
            e.start()

        mock_node.run.assert_called_once()



from __future__ import annotations

from typing import Any

from goalee.entity import Entity
from goalee.goal import Goal, GoalState
from goalee.types import Orientation, Point


class PoseGoal(Goal):
    def __init__(
        self,
        entity: Entity,
        position: Point,
        orientation: Orientation,
        deviation_pos: float = 0.0,
        deviation_ori: float = 0.0,
        name: str | None = None,
        event_emitter: Any | None = None,
        max_duration: float | None = None,
        min_duration: float | None = None,
    ):
        super().__init__(
            [entity],
            event_emitter,
            name=name,
            max_duration=max_duration,
            min_duration=min_duration,
        )
        self._entity = entity
        self._position = position
        self._orientation = orientation
        self._deviation_pos = deviation_pos
        self._deviation_ori = deviation_ori

    def on_enter(self):
        self.log_debug(
            f"Starting PoseGoal <{self._name}> with params:\n"
            f"-> Entity: {self._entity.name}\n"
            f"-> Position: {self._position}\n"
            f"-> Orientation: {self._orientation}\n"
            f"-> Deviation Position: {self._deviation_pos}\n"
            f"-> Deviation Orientation: {self._deviation_ori}"
        )

    def check_pose(self):
        if self._last_state.get("position", None) is None:
            return
        pos = self._last_state["position"]
        if pos.get("x", "None") is None or pos.get("y", "None") is None:
            return
        if self._last_state.get("orientation", None) is None:
            return
        ori = self._last_state["orientation"]
        if ori.get("x", "None") is None or ori.get("y", "None") is None:
            return
        reached = (
            pos > (self._position - self._deviation_pos)
            and pos < (self._position + self._deviation_pos)
            and ori > (self._orientation - self._deviation_ori)
            and ori < (self._orientation + self._deviation_ori)
        )
        if reached:
            self.set_state(GoalState.COMPLETED)

    def tick(self):
        self._last_state = self._entity.attributes.copy()
        self.check_pose()


class PositionGoal(Goal):
    def __init__(
        self,
        entity: Entity,
        position: Point,
        deviation: float = 0.0,
        name: str | None = None,
        event_emitter: Any | None = None,
        max_duration: float | None = None,
        min_duration: float | None = None,
    ):
        super().__init__(
            [entity],
            event_emitter,
            name=name,
            max_duration=max_duration,
            min_duration=min_duration,
        )
        self._entity = entity
        self._position = position
        self._deviation = deviation

    def on_enter(self):
        self.log_debug(
            f"Starting PositionGoal <{self._name}> with params:\n"
            f"-> Entity: {self._entity.name}\n"
            f"-> Position: {self._position}\n"
            f"-> Deviation: {self._deviation}\n"
            f"-> Max Duration: {self._max_duration}\n"
            f"-> Min Duration: {self._min_duration}"
        )

    def check_pos(self):
        if self._last_state.get("position", None) is None:
            return
        pos = self._last_state["position"]
        if (
            pos.get("x", "None") is None
            and pos.get("y", "None") is None
            and pos.get("z", "None") is None
        ):
            self.log_warning("Received invalid position values for x,y,z")
            return
        reached = (
            pos["x"] > (self._position.x - self._deviation)
            and pos["x"] < (self._position.x + self._deviation)
            and pos["y"] > (self._position.y - self._deviation)
            and pos["y"] < (self._position.y + self._deviation)
            and pos["z"] > (self._position.z - self._deviation)
            and pos["z"] < (self._position.z + self._deviation)
        )
        if reached:
            self.set_state(GoalState.COMPLETED)

    def tick(self):
        self._last_state = self._entity.attributes.copy()
        self.check_pos()


class OrientationGoal(Goal):
    def __init__(
        self,
        entity: Entity,
        orientation: Orientation,
        deviation: float = 0.0,
        name: str | None = None,
        event_emitter: Any | None = None,
        max_duration: float | None = None,
        min_duration: float | None = None,
    ):
        super().__init__(
            [entity],
            event_emitter,
            name=name,
            max_duration=max_duration,
            min_duration=min_duration,
        )
        self._entity = entity
        self._orientation = orientation
        self._deviation = deviation

    def on_enter(self):
        self.log_debug(
            f"Starting OrientationGoal <{self._name}> with params:\n"
            f"-> Entity: {self._entity.name}\n"
            f"-> Orientation: {self._orientation}\n"
            f"-> Deviation: {self._deviation}\n"
            f"-> Max Duration: {self._max_duration}\n"
            f"-> Min Duration: {self._min_duration}"
        )

    def check_ori(self):
        if self._last_state.get("orientation", None) is None:
            return
        ori = self._last_state["orientation"]
        if (
            ori.get("roll", None) is None
            and ori.get("pitch", None) is None
            and ori.get("yaw", None) is None
        ):
            return
        reached = (
            ori["roll"] > (self._orientation.roll - self._deviation)
            and ori["roll"] < (self._orientation.roll + self._deviation)
            and ori["pitch"] > (self._orientation.pitch - self._deviation)
            and ori["pitch"] < (self._orientation.pitch + self._deviation)
            and ori["yaw"] > (self._orientation.yaw - self._deviation)
            and ori["yaw"] < (self._orientation.yaw + self._deviation)
        )
        if reached:
            self.set_state(GoalState.COMPLETED)

    def tick(self):
        self._last_state = self._entity.attributes.copy()
        self.check_ori()



from __future__ import annotations

from typing import Any

from goalee.entity import Entity
from goalee.goal import Goal, GoalState
from goalee.types import Point


class WaypointTrajectoryGoal(Goal):
    def __init__(
        self,
        entity: Entity,
        waypoints: list[Point],
        deviation: float = 0.0,
        name: str | None = None,
        event_emitter: Any | None = None,
        max_duration: float | None = None,
        min_duration: float | None = None,
    ):
        super().__init__(
            [entity],
            event_emitter,
            name=name,
            max_duration=max_duration,
            min_duration=min_duration,
        )
        self._entity = entity
        self._waypoints = waypoints
        self._deviation = deviation
        self._waypoints_reached_map = [False] * len(waypoints)

    def on_enter(self):
        self.log_debug(
            f"Starting PositionGoal <{self._name}> with params:\n"
            f"-> Entity: {self._entity.name}\n"
            f"-> Waypoints: {self._waypoints}\n"
            f"-> Deviation: {self._deviation}\n"
            f"-> Max Duration: {self._max_duration}\n"
            f"-> Min Duration: {self._min_duration}"
        )

    def current_target_waypoint(self):
        for i, reached in enumerate(self._waypoints_reached_map):
            if not reached:
                return self._waypoints[i], i

    def check_reached_waypoint(self):
        if self._last_state.get("position", None) is None:
            return
        pos = self._last_state["position"]
        if (
            pos.get("x", "None") is None
            and pos.get("y", "None") is None
            and pos.get("z", "None") is None
        ):
            self.log_warning("Received invalid position values for x,y,z")
            return
        current_target, idx = self.current_target_waypoint()
        reached = (
            pos["x"] > (current_target.x - self._deviation)
            and pos["x"] < (current_target.x + self._deviation)
            and pos["y"] > (current_target.y - self._deviation)
            and pos["y"] < (current_target.y + self._deviation)
            and pos["z"] > (current_target.z - self._deviation)
            and pos["z"] < (current_target.z + self._deviation)
        )
        if reached:
            self.log_info(f"Reached waypoint {idx}")
            self._waypoints_reached_map[idx] = True

    def tick(self):
        self._last_state = self._entity.attributes.copy()
        self.check_reached_waypoint()
        if all(self._waypoints_reached_map):
            self.set_state(GoalState.COMPLETED)



from __future__ import annotations

from typing import Any

from goalee.goal import Goal, GoalState
from goalee.rtmonitor import RTMonitor


class GoalRepeater(Goal):
    def __init__(
        self,
        goal: Goal,
        times: int,
        name: str | None = None,
        event_emitter: Any | None = None,
        max_duration: float | None = None,
        min_duration: float | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            entities=None,
            event_emitter=event_emitter,
            name=name,
            max_duration=max_duration,
            min_duration=min_duration,
            *args,
            **kwargs,
        )
        self._goal = goal

        self._repeat_times = times
        self._times = 0

        if (
            self._goal._max_duration is None
            or self._goal._max_duration > self._max_duration
        ) and self._max_duration is not None:
            self._goal._max_duration = self._max_duration

    def set_tick_freq(self, freq: int):
        self._goal.set_tick_freq(freq)

    def serialize(self):
        return {
            **super().serialize(),
            "times": self._repeat_times,
            "goals": [self._goal.serialize()],
        }

    def on_enter(self):
        self.log_info(
            f"Starting Goal-Repeater <{self._name}>:\n"
            f"Parameters:\n"
            f"  Goal: {self._goal.__class__.__name__}:{self._goal.name}\n"
            f"  Times: {self._repeat_times}\n"
            f"  Max Duration: {self._max_duration}\n"
            f"  Min Duration: {self._min_duration}"
        )

    def enter(self, rtmonitor: RTMonitor = None):
        self._ts_start = self.get_current_ts()
        self.set_state(GoalState.RUNNING)
        self.on_enter()
        _states = []
        _durations = []
        while self._times < self._repeat_times and self._state not in (
            GoalState.TERMINATED,
            GoalState.FAILED,
            GoalState.COMPLETED,
        ):
            self._times += 1
            self._goal.enter()
            if self._state != GoalState.RUNNING:
                break
            _states.append(self._goal.state)
            _durations.append(self._goal.duration)
            if (
                self._max_duration not in (None, 0)
                and self.get_current_elapsed() > self._max_duration
            ):
                break
            self._goal.reset()
        elapsed = self.get_current_elapsed()
        if (
            self._max_duration not in (None, 0)
            and elapsed > self._max_duration
            or self._min_duration not in (None, 0)
            and elapsed < self._min_duration
        ):
            self.set_state(GoalState.FAILED)
        elif self._state == GoalState.RUNNING:
            self.set_state(
                GoalState.COMPLETED
                if all([s == GoalState.COMPLETED for s in _states])
                else GoalState.FAILED
            )
        else:  # Terminated / Failed
            pass
        self.on_exit()
        return self



from __future__ import annotations

import time
from unittest.mock import MagicMock

import pytest

from goalee.entity import Entity
from goalee.goal import Goal, GoalState


class MockGoal(Goal):
    def __init__(self, complete_after_ticks=1, **kwargs):
        super().__init__(**kwargs)
        self._tick_count = 0
        self._complete_after = complete_after_ticks

    def on_enter(self):
        pass

    def tick(self):
        self._tick_count += 1
        if self._tick_count >= self._complete_after:
            self.set_state(GoalState.COMPLETED)


class TestGoalState:
    def test_enum_values(self):
        assert GoalState.IDLE == 0
        assert GoalState.RUNNING == 1
        assert GoalState.COMPLETED == 2
        assert GoalState.FAILED == 3
        assert GoalState.TERMINATED == 4

    def test_enum_names(self):
        assert GoalState.IDLE.name == "IDLE"
        assert GoalState.COMPLETED.name == "COMPLETED"


class TestGoalCreation:
    def test_defaults(self):
        g = MockGoal()
        assert g.state == GoalState.IDLE
        assert g.entities == []
        assert g.name is not None
        assert len(g.name) > 0
        assert g._max_duration is None
        assert g._min_duration is None
        assert g._for_duration is None
        assert g._freq == 10

    def test_explicit_name(self):
        g = MockGoal(name="my_goal")
        assert g.name == "my_goal"

    def test_empty_name_generates_random(self):
        g = MockGoal(name="")
        assert len(g.name) > 0
        assert g.name != ""

    def test_none_name_generates_random(self):
        g = MockGoal(name=None)
        assert len(g.name) > 0

    def test_with_entities(self):
        e = Entity(name="e1", etype="sensor", topic="t", attributes=["a"])
        g = MockGoal(entities=[e])
        assert len(g.entities) == 1
        assert g.entities[0].name == "e1"

    def test_with_durations(self):
        g = MockGoal(max_duration=10.0, min_duration=1.0, for_duration=5.0)
        assert g._max_duration == 10.0
        assert g._min_duration == 1.0
        assert g._for_duration == 5.0

    def test_with_tick_freq(self):
        g = MockGoal(tick_freq=20)
        assert g._freq == 20


class TestGoalSetState:
    def test_valid_states(self):
        g = MockGoal()
        for state in GoalState:
            g._state = GoalState.IDLE  # reset to avoid same-state no-op
            g.set_state(state)
            assert g.state == state

    def test_invalid_state_raises(self):
        g = MockGoal()
        with pytest.raises((ValueError, TypeError)):
            g.set_state(99)

    def test_same_state_noop(self):
        g = MockGoal()
        assert g.state == GoalState.IDLE
        g.set_state(GoalState.IDLE)
        assert g.state == GoalState.IDLE


class TestGoalProperties:
    def test_name(self):
        g = MockGoal(name="test")
        assert g.name == "test"

    def test_state(self):
        g = MockGoal()
        assert g.state == GoalState.IDLE

    def test_status_false_when_idle(self):
        g = MockGoal()
        assert g.status is False

    def test_status_true_when_completed(self):
        g = MockGoal()
        g.set_state(GoalState.RUNNING)
        g.set_state(GoalState.COMPLETED)
        assert g.status is True

    def test_status_false_when_failed(self):
        g = MockGoal()
        g.set_state(GoalState.RUNNING)
        g.set_state(GoalState.FAILED)
        assert g.status is False

    def test_entities(self):
        g = MockGoal()
        assert g.entities == []

    def test_duration(self):
        g = MockGoal()
        assert g.duration == 0.0  # _ts_exit(-1) - _ts_start(-1) = 0


class TestGoalSetTickFreq:
    def test_set_tick_freq(self):
        g = MockGoal()
        g.set_tick_freq(50)
        assert g._freq == 50


class TestGoalSerialize:
    def test_serialize_structure(self):
        e = Entity(name="e1", etype="sensor", topic="t", attributes=["a"])
        g = MockGoal(
            name="test_goal",
            entities=[e],
            max_duration=10.0,
            min_duration=1.0,
            for_duration=5.0,
        )
        result = g.serialize()
        assert result["name"] == "test_goal"
        assert result["type"] == "MockGoal"
        assert result["state"] == "IDLE"
        assert result["max_duration"] == 10.0
        assert result["min_duration"] == 1.0
        assert result["for_duration"] == 5.0
        assert result["entities"] == ["e1"]
        assert "elapsed" in result
        assert "ts_start" in result
        assert "ts_exit" in result


class TestGoalEnter:
    def test_enter_completes(self):
        g = MockGoal(complete_after_ticks=1, name="enter_test")
        result = g.enter()
        assert result is g
        assert g.state == GoalState.COMPLETED
        assert g._ts_start > 0
        assert g._ts_exit > 0

    def test_enter_with_rtmonitor(self):
        mock_rtm = MagicMock()
        g = MockGoal(complete_after_ticks=1, name="rtm_test")
        g.enter(rtmonitor=mock_rtm)
        assert g._rtmonitor is mock_rtm


class TestGoalRunUntilExit:
    def test_completes_on_tick(self):
        g = MockGoal(complete_after_ticks=2, name="run_test")
        g._ts_start = time.time()
        g.set_state(GoalState.RUNNING)
        g.run_until_exit()
        assert g.state == GoalState.COMPLETED
        assert g._tick_count == 2

    def test_fails_on_max_duration(self):
        class NeverCompleteGoal(Goal):
            def on_enter(self):
                pass

            def tick(self):
                pass

        g = NeverCompleteGoal(name="timeout_test", max_duration=0.001, tick_freq=1000)
        g._ts_start = time.time() - 1.0  # started 1 second ago
        g.set_state(GoalState.RUNNING)
        g.run_until_exit()
        assert g.state == GoalState.FAILED

    def test_fails_on_min_duration(self):
        g = MockGoal(complete_after_ticks=1, name="min_dur_test", min_duration=999.0)
        g._ts_start = time.time()
        g.set_state(GoalState.RUNNING)
        g.run_until_exit()
        assert g.state == GoalState.FAILED

    def test_no_max_duration_continues(self):
        g = MockGoal(complete_after_ticks=3, name="no_max_test", max_duration=None)
        g._ts_start = time.time()
        g.set_state(GoalState.RUNNING)
        g.run_until_exit()
        assert g.state == GoalState.COMPLETED
        assert g._tick_count == 3

    def test_zero_max_duration_continues(self):
        g = MockGoal(complete_after_ticks=2, name="zero_max_test", max_duration=0)
        g._ts_start = time.time()
        g.set_state(GoalState.RUNNING)
        g.run_until_exit()
        assert g.state == GoalState.COMPLETED

    def test_terminated_exits_loop(self):
        class TerminateGoal(Goal):
            def __init__(self, **kwargs):
                super().__init__(**kwargs)
                self._count = 0

            def on_enter(self):
                pass

            def tick(self):
                self._count += 1
                if self._count >= 2:
                    self.set_state(GoalState.TERMINATED)

        g = TerminateGoal(name="term_test")
        g._ts_start = time.time()
        g.set_state(GoalState.RUNNING)
        g.run_until_exit()
        assert g.state == GoalState.TERMINATED

    def test_min_duration_none_no_fail(self):
        g = MockGoal(complete_after_ticks=1, name="min_none", min_duration=None)
        g._ts_start = time.time()
        g.set_state(GoalState.RUNNING)
        g.run_until_exit()
        assert g.state == GoalState.COMPLETED

    def test_min_duration_zero_no_fail(self):
        g = MockGoal(complete_after_ticks=1, name="min_zero", min_duration=0)
        g._ts_start = time.time()
        g.set_state(GoalState.RUNNING)
        g.run_until_exit()
        assert g.state == GoalState.COMPLETED


class TestGoalTerminate:
    def test_terminate(self):
        g = MockGoal(name="term")
        g.set_state(GoalState.RUNNING)
        g.terminate()
        assert g.state == GoalState.TERMINATED


class TestGoalReset:
    def test_reset(self):
        g = MockGoal(name="reset_test")
        g.set_state(GoalState.RUNNING)
        g._ts_start = 100.0
        g._ts_hold = 50.0
        g._ts_exit = 200.0
        g._duration = 100.0
        g.reset()
        assert g.state == GoalState.IDLE
        assert g._ts_start == -1.0
        assert g._ts_hold == -1.0
        assert g._ts_exit == -1.0
        assert g._duration == -1.0


class TestGoalLogging:
    def test_log_namespace(self):
        g = MockGoal(name="mygoal")
        assert g.log_namespace() == "MockGoal:mygoal"

    def test_log_info(self):
        g = MockGoal(name="logtest")
        g.log_info("test message")

    def test_log_warning(self):
        g = MockGoal(name="logtest")
        g.log_warning("test warning")

    def test_log_error(self):
        g = MockGoal(name="logtest")
        g.log_error("test error")

    def test_log_debug(self):
        g = MockGoal(name="logtest")
        g.log_debug("test debug")

    def test_log_returns_logger(self):
        g = MockGoal(name="logtest")
        assert g.log() is not None


class TestGoalTimestamps:
    def test_get_current_ts(self):
        g = MockGoal(name="ts_test")
        ts = g.get_current_ts()
        assert isinstance(ts, float)
        assert ts > 0

    def test_get_current_ts_ms(self):
        g = MockGoal(name="ts_ms_test")
        ts_ms = g.get_current_ts_ms()
        assert isinstance(ts_ms, int)
        assert ts_ms > 0

    def test_get_current_elapsed(self):
        g = MockGoal(name="elapsed_test")
        g._ts_start = time.time() - 1.0
        elapsed = g.get_current_elapsed()
        assert elapsed >= 0.9


class TestGoalAbstractMethods:
    def test_on_enter_raises(self):
        g = Goal.__new__(Goal)
        with pytest.raises(NotImplementedError):
            g.on_enter()

    def test_tick_raises(self):
        g = Goal.__new__(Goal)
        with pytest.raises(NotImplementedError):
            g.tick()


class TestGoalGenRandomName:
    def test_generates_unique_names(self):
        g = MockGoal()
        name1 = g._gen_random_name()
        name2 = g._gen_random_name()
        assert name1 != name2
        assert isinstance(name1, str)
        assert len(name1) > 0
        assert "-" not in name1


class TestGoalStateChangeEvent:
    def test_send_state_change_event_with_rtmonitor(self):
        mock_rtm = MagicMock()
        g = MockGoal(name="event_test")
        g._rtmonitor = mock_rtm
        g._ts_start = time.time()
        g.set_state(GoalState.RUNNING)
        assert mock_rtm.send_event.called

    def test_no_event_without_rtmonitor(self):
        g = MockGoal(name="no_rtm")
        g.set_state(GoalState.RUNNING)


class TestGoalReportState:
    def test_report_state_no_crash(self):
        g = MockGoal(name="report_test")
        g._report_state()


class TestGoalOnExit:
    def test_on_exit_sets_ts_exit(self):
        g = MockGoal(name="exit_test")
        g.on_exit()
        assert g._ts_exit > 0

    def test_on_reset_default(self):
        g = MockGoal(name="reset_default")
        g.on_reset()


class TestGoalSetRtmonitor:
    def test_set_rtmonitor(self):
        g = MockGoal(name="rtm_set")
        mock_rtm = MagicMock()
        g.set_rtmonitor(mock_rtm)
        assert g._rtmonitor is mock_rtm


class TestGoalSendStateChangeEventDuration:
    def test_positive_duration(self):
        mock_rtm = MagicMock()
        g = MockGoal(name="dur_pos")
        g._rtmonitor = mock_rtm
        g._ts_start = 100.0
        g._ts_exit = 110.0
        g._state = GoalState.RUNNING
        g.set_state(GoalState.COMPLETED)
        call_args = mock_rtm.send_event.call_args
        event = call_args[0][0]
        assert event.data["duration"] == 10.0

    def test_negative_duration_uses_elapsed(self):
        mock_rtm = MagicMock()
        g = MockGoal(name="dur_neg")
        g._rtmonitor = mock_rtm
        g._ts_start = time.time()
        g._ts_exit = -1.0
        g._state = GoalState.IDLE
        g.set_state(GoalState.RUNNING)
        call_args = mock_rtm.send_event.call_args
        event = call_args[0][0]
        assert event.data["duration"] >= 0



from __future__ import annotations

import os

ZERO_LOGS = int(os.getenv("GOALDSL_ZERO_LOGS", 0))
LOG_LEVEL = os.getenv("GOALDSL_LOG_LEVEL", "INFO")
GOAL_TICK_FREQ_HZ = int(os.getenv("GOAL_TICK_FREQ_HZ", 10))



from __future__ import annotations

from unittest.mock import MagicMock, patch

import goalee
from goalee import AMQPBroker, Broker, Entity, MQTTBroker, RedisBroker, Scenario
from goalee.complex_goal import ComplexGoalAlgorithm
from goalee.goal import GoalState
from goalee.types import Orientation, Point


def test_version():
    assert isinstance(goalee.__version__, str)
    parts = goalee.__version__.split(".")
    assert len(parts) == 3
    assert all(p.isdigit() for p in parts)


def test_public_imports():
    assert Scenario is not None
    assert Entity is not None
    assert Broker is not None
    assert RedisBroker is not None
    assert MQTTBroker is not None
    assert AMQPBroker is not None


def test_goal_state_enum():
    assert GoalState.IDLE == 0
    assert GoalState.RUNNING == 1
    assert GoalState.COMPLETED == 2
    assert GoalState.FAILED == 3
    assert GoalState.TERMINATED == 4


def test_broker_defaults():
    redis = RedisBroker()
    assert redis.host == "localhost"
    assert redis.port == 6379
    assert redis.db == 0

    mqtt = MQTTBroker()
    assert mqtt.host == "localhost"
    assert mqtt.port == 1883

    amqp = AMQPBroker()
    assert amqp.host == "localhost"
    assert amqp.port == 5672
    assert amqp.vhost == "/"


def test_complex_goal_algorithm_enum():
    assert ComplexGoalAlgorithm.ALL_ACCOMPLISHED == 0
    assert ComplexGoalAlgorithm.NONE_ACCOMPLISHED == 2
    assert ComplexGoalAlgorithm.AT_LEAST_ONE_ACCOMPLISHED == 3


def test_point_operations():
    p1 = Point(1.0, 2.0, 3.0)
    p2 = Point(4.0, 5.0, 6.0)
    result = p1 + p2
    assert result.x == 5.0
    assert result.y == 7.0
    assert result.z == 9.0

    diff = p2 - p1
    assert diff.x == 3.0
    assert diff.y == 3.0
    assert diff.z == 3.0


def test_orientation_operations():
    o1 = Orientation(0.1, 0.2, 0.3)
    o2 = Orientation(0.4, 0.5, 0.6)
    result = o1 + o2
    assert abs(result.roll - 0.5) < 1e-9
    assert abs(result.pitch - 0.7) < 1e-9
    assert abs(result.yaw - 0.9) < 1e-9


def test_scenario_creation():
    broker = RedisBroker()
    with patch.object(Scenario, "_create_comm_node", return_value=MagicMock()):
        scenario = Scenario(name="test_scenario", broker=broker)
    assert scenario.name == "test_scenario"


def test_entity_creation():
    broker = RedisBroker()
    entity = Entity(
        name="test_entity",
        etype="sensor",
        topic="test.topic",
        attributes=["temp", "humidity"],
        source=broker,
    )
    assert entity.name == "test_entity"
    assert entity.topic == "test.topic"
    assert entity.attributes["temp"] is None
    assert entity.attributes["humidity"] is None



"""Timing goal types for runtime verification of temporal properties."""

from __future__ import annotations

import statistics
import time
from collections import deque
from datetime import datetime
from typing import Any, Callable

from goalee.entity import Entity
from goalee.goal import Goal, GoalState


class RateGoal(Goal):
    """Verifies entity publishes at expected interval ± tolerance."""

    def __init__(
        self,
        entity: Entity,
        name: str | None = None,
        interval: float | None = None,
        tolerance: float | None = None,
        jitter: float | None = None,
        window: int = 20,
        event_emitter: Any | None = None,
        max_duration: float | None = None,
        min_duration: float | None = None,
        for_duration: float | None = None,
    ):
        super().__init__(
            [entity],
            event_emitter,
            name=name,
            max_duration=max_duration,
            min_duration=min_duration,
            for_duration=for_duration,
        )
        self.entity = entity
        self._interval = interval  # expected interval in seconds
        self._tolerance = tolerance  # acceptable deviation in seconds
        self._jitter = jitter  # max acceptable jitter (std of intervals)
        self._window = window
        self._timestamps: deque = deque(maxlen=window)
        self._last_state: dict | None = None

    def on_enter(self):
        self.log_debug(
            f"Starting RateGoal <{self.name}>:\n"
            f"Entity: {self.entity.name}\n"
            f"Interval: {self._interval}s\n"
            f"Tolerance: {self._tolerance}s\n"
            f"Window: {self._window}"
        )
        self._last_state = None

    def tick(self):
        current_state = self.entity.attributes.copy()
        if self._last_state is None:
            self._last_state = current_state
            self._timestamps.append(time.time())
            return

        if current_state != self._last_state:
            self._timestamps.append(time.time())
            self._last_state = current_state

        if len(self._timestamps) < 2:
            return

        # Calculate intervals between consecutive timestamps
        intervals = [
            self._timestamps[i + 1] - self._timestamps[i]
            for i in range(len(self._timestamps) - 1)
        ]

        if not intervals:
            return

        avg_interval = statistics.mean(intervals)

        # Check if average interval is within tolerance of expected
        if self._tolerance is not None and self._interval is not None:
            if abs(avg_interval - self._interval) <= self._tolerance:
                # Check jitter if specified
                if self._jitter is not None and len(intervals) > 1:
                    interval_jitter = statistics.stdev(intervals)
                    if interval_jitter <= self._jitter:
                        self.set_state(GoalState.COMPLETED)
                else:
                    self.set_state(GoalState.COMPLETED)


class LatencyGoal(Goal):
    """Measures time between trigger and response events."""

    def __init__(
        self,
        trigger_entities: list[Entity] | None = None,
        response_entities: list[Entity] | None = None,
        name: str | None = None,
        trigger_condition: Callable | None = None,
        response_condition: Callable | None = None,
        max_latency: float | None = None,
        event_emitter: Any | None = None,
        max_duration: float | None = None,
        min_duration: float | None = None,
        for_duration: float | None = None,
    ):
        all_entities = list(set((trigger_entities or []) + (response_entities or [])))
        super().__init__(
            all_entities,
            event_emitter,
            name=name,
            max_duration=max_duration,
            min_duration=min_duration,
            for_duration=for_duration,
        )
        self._trigger_entities = trigger_entities or []
        self._response_entities = response_entities or []
        self._trigger_condition = trigger_condition
        self._response_condition = response_condition
        self._max_latency = max_latency
        self._trigger_ts: float | None = None

    def on_enter(self):
        self.log_debug(
            f"Starting LatencyGoal <{self.name}>:\nMax Latency: {self._max_latency}s"
        )
        self._trigger_ts = None

    def _get_entities_map(self, entities):
        return {e.name: e for e in entities}

    def tick(self):
        try:
            # Check trigger
            if self._trigger_ts is None:
                if self._trigger_condition is not None:
                    e_map = self._get_entities_map(self._trigger_entities)
                    if callable(self._trigger_condition):
                        if self._trigger_condition(e_map):
                            self._trigger_ts = time.time()
                else:
                    # Entity-based trigger: any state change
                    self._trigger_ts = time.time()
                return

            # Check response
            if self._response_condition is not None:
                e_map = self._get_entities_map(self._response_entities)
                if callable(self._response_condition):
                    if self._response_condition(e_map):
                        latency = time.time() - self._trigger_ts
                        if (
                            self._max_latency is not None
                            and latency <= self._max_latency
                        ):
                            self.set_state(GoalState.COMPLETED)
                        else:
                            self.set_state(GoalState.FAILED)
            else:
                # Entity-based response: any state change
                latency = time.time() - self._trigger_ts
                if self._max_latency is not None and latency <= self._max_latency:
                    self.set_state(GoalState.COMPLETED)
                else:
                    self.set_state(GoalState.FAILED)
        except TypeError:
            self.log_debug("Condition evaluation skipped (entity not yet initialized)")


class OrderingGoal(Goal):
    """Verifies goals complete in specified sequence within time window."""

    def __init__(
        self,
        goals: list[str] | None = None,
        name: str | None = None,
        within: float | None = None,
        event_emitter: Any | None = None,
        max_duration: float | None = None,
        min_duration: float | None = None,
        for_duration: float | None = None,
    ):
        super().__init__(
            [],
            event_emitter,
            name=name,
            max_duration=max_duration,
            min_duration=min_duration,
            for_duration=for_duration,
        )
        self._goal_names = goals or []
        self._within = within
        self._completed_order: list[str] = []

    def on_enter(self):
        self.log_debug(
            f"Starting OrderingGoal <{self.name}>:\n"
            f"Sequence: {self._goal_names}\n"
            f"Within: {self._within}s"
        )
        self._completed_order = []

    def tick(self):
        # In a real scenario, this would check goal completion order
        # via the scenario's goal registry. Placeholder for now.
        pass


class DeadlineGoal(Goal):
    """Condition must be met by absolute wall-clock time."""

    def __init__(
        self,
        entities: list[Entity] | None = None,
        name: str | None = None,
        condition: Callable | None = None,
        deadline_time: str | None = None,
        after_time: str | None = None,
        event_emitter: Any | None = None,
        max_duration: float | None = None,
        min_duration: float | None = None,
        for_duration: float | None = None,
    ):
        super().__init__(
            entities,
            event_emitter,
            name=name,
            max_duration=max_duration,
            min_duration=min_duration,
            for_duration=for_duration,
        )
        self._condition = condition
        self._deadline_time = deadline_time
        self._after_time = after_time

    def on_enter(self):
        self.log_debug(
            f"Starting DeadlineGoal <{self.name}>:\n"
            f"Deadline: {self._deadline_time}\n"
            f"After: {self._after_time}"
        )

    def _parse_time(self, time_str: str) -> datetime:
        parts = time_str.split(":")
        now = datetime.now()
        return now.replace(
            hour=int(parts[0]),
            minute=int(parts[1]) if len(parts) > 1 else 0,
            second=int(parts[2]) if len(parts) > 2 else 0,
            microsecond=0,
        )

    def _get_entities_map(self):
        return {e.name: e for e in self._entities}

    def tick(self):
        now = datetime.now()

        # Check if we're past the after time (if specified)
        if self._after_time is not None:
            after_dt = self._parse_time(self._after_time)
            if now < after_dt:
                return

        # Check deadline
        if self._deadline_time is not None:
            deadline_dt = self._parse_time(self._deadline_time)
            if now > deadline_dt:
                self.set_state(GoalState.FAILED)
                return

        # Evaluate condition
        try:
            if self._condition is not None:
                if callable(self._condition):
                    e_map = self._get_entities_map()
                    if self._condition(e_map):
                        self.set_state(GoalState.COMPLETED)
        except TypeError:
            self.log_debug("Condition evaluation skipped (entity not yet initialized)")



from __future__ import annotations

from unittest.mock import patch

import pytest

from goalee.complex_goal import ComplexGoal, ComplexGoalAlgorithm
from goalee.goal import Goal, GoalState


class MockGoal(Goal):
    def __init__(self, complete_on_enter=True, fail_on_enter=False, **kwargs):
        super().__init__(**kwargs)
        self._complete_on_enter = complete_on_enter
        self._fail_on_enter = fail_on_enter

    def on_enter(self):
        pass

    def tick(self):
        if self._complete_on_enter:
            self.set_state(GoalState.COMPLETED)
        elif self._fail_on_enter:
            self.set_state(GoalState.FAILED)


# --- Construction ---


def test_default_algorithm():
    cg = ComplexGoal(name="cg")
    assert cg._algorithm == ComplexGoalAlgorithm.ALL_ACCOMPLISHED
    assert cg._goals == []
    assert cg._x_accomplished is None


def test_explicit_algorithm_and_accomplished():
    cg = ComplexGoal(
        name="cg",
        algorithm=ComplexGoalAlgorithm.EXACTLY_X_ACCOMPLISHED,
        accomplished=2,
    )
    assert cg._algorithm == ComplexGoalAlgorithm.EXACTLY_X_ACCOMPLISHED
    assert cg._x_accomplished == 2


def test_goals_property():
    cg = ComplexGoal(name="cg")
    g1 = MockGoal(name="g1")
    cg._goals.append(g1)
    assert cg.goals == [g1]


# --- add_goal ---


def test_add_goal_appends():
    cg = ComplexGoal(name="cg")
    g1 = MockGoal(name="g1")
    g2 = MockGoal(name="g2")
    cg.add_goal(g1)
    cg.add_goal(g2)
    assert cg.goals == [g1, g2]


def test_add_goal_propagates_max_duration_when_child_none():
    cg = ComplexGoal(name="cg", max_duration=5.0)
    g = MockGoal(name="g1", max_duration=None)
    cg.add_goal(g)
    assert g._max_duration == 5.0


def test_add_goal_propagates_max_duration_when_child_larger():
    cg = ComplexGoal(name="cg", max_duration=5.0)
    g = MockGoal(name="g1", max_duration=10.0)
    cg.add_goal(g)
    assert g._max_duration == 5.0


def test_add_goal_keeps_child_max_duration_when_smaller():
    cg = ComplexGoal(name="cg", max_duration=10.0)
    g = MockGoal(name="g1", max_duration=3.0)
    cg.add_goal(g)
    assert g._max_duration == 3.0


def test_add_goal_no_propagation_when_parent_max_none():
    cg = ComplexGoal(name="cg", max_duration=None)
    g = MockGoal(name="g1", max_duration=5.0)
    cg.add_goal(g)
    assert g._max_duration == 5.0


def test_add_goal_propagates_min_duration_when_child_none():
    cg = ComplexGoal(name="cg", min_duration=2.0)
    g = MockGoal(name="g1", min_duration=None)
    cg.add_goal(g)
    assert g._min_duration == 2.0


def test_add_goal_propagates_min_duration_when_child_smaller():
    cg = ComplexGoal(name="cg", min_duration=5.0)
    g = MockGoal(name="g1", min_duration=2.0)
    cg.add_goal(g)
    assert g._min_duration == 5.0


def test_add_goal_keeps_child_min_duration_when_larger():
    cg = ComplexGoal(name="cg", min_duration=2.0)
    g = MockGoal(name="g1", min_duration=10.0)
    cg.add_goal(g)
    assert g._min_duration == 10.0


def test_add_goal_no_propagation_when_parent_min_none():
    cg = ComplexGoal(name="cg", min_duration=None)
    g = MockGoal(name="g1", min_duration=3.0)
    cg.add_goal(g)
    assert g._min_duration == 3.0


# --- set_state_cascaded ---


def test_set_state_cascaded():
    cg = ComplexGoal(name="cg")
    g1 = MockGoal(name="g1")
    g2 = MockGoal(name="g2")
    cg.add_goal(g1)
    cg.add_goal(g2)
    cg.set_state_cascaded(GoalState.TERMINATED)
    assert g1.state == GoalState.TERMINATED
    assert g2.state == GoalState.TERMINATED
    assert cg.state == GoalState.TERMINATED


# --- set_tick_freq ---


def test_set_tick_freq_propagates():
    cg = ComplexGoal(name="cg")
    g1 = MockGoal(name="g1")
    g2 = MockGoal(name="g2")
    cg.add_goal(g1)
    cg.add_goal(g2)
    cg.set_tick_freq(50)
    assert g1._freq == 50
    assert g2._freq == 50


# --- serialize ---


def test_serialize():
    cg = ComplexGoal(name="cg", algorithm=ComplexGoalAlgorithm.NONE_ACCOMPLISHED)
    g1 = MockGoal(name="g1")
    cg.add_goal(g1)
    data = cg.serialize()
    assert data["algorithm"] == "NONE_ACCOMPLISHED"
    assert data["name"] == "cg"
    assert len(data["goals"]) == 1
    assert data["goals"][0]["name"] == "g1"


# --- reset ---


def test_reset():
    cg = ComplexGoal(name="cg")
    g1 = MockGoal(name="g1", complete_on_enter=True)
    g2 = MockGoal(name="g2", complete_on_enter=True)
    cg.add_goal(g1)
    cg.add_goal(g2)
    cg.enter()
    assert cg.state == GoalState.COMPLETED
    cg.reset()
    assert cg.state == GoalState.IDLE
    assert g1.state == GoalState.IDLE
    assert g2.state == GoalState.IDLE
    assert cg._ts_start == -1.0
    assert cg._ts_exit == -1.0
    assert cg._duration == -1.0


# --- ALL_ACCOMPLISHED ---


def test_all_accomplished_all_complete():
    cg = ComplexGoal(name="cg", algorithm=ComplexGoalAlgorithm.ALL_ACCOMPLISHED)
    cg.add_goal(MockGoal(complete_on_enter=True, name="g1"))
    cg.add_goal(MockGoal(complete_on_enter=True, name="g2"))
    cg.enter()
    assert cg.state == GoalState.COMPLETED


def test_all_accomplished_one_fails():
    cg = ComplexGoal(name="cg", algorithm=ComplexGoalAlgorithm.ALL_ACCOMPLISHED)
    cg.add_goal(MockGoal(complete_on_enter=True, name="g1"))
    cg.add_goal(
        MockGoal(
            complete_on_enter=False, fail_on_enter=True, name="g2", max_duration=0.01
        )
    )
    cg.enter()
    assert cg.state == GoalState.FAILED


def test_all_accomplished_all_fail():
    cg = ComplexGoal(name="cg", algorithm=ComplexGoalAlgorithm.ALL_ACCOMPLISHED)
    cg.add_goal(
        MockGoal(
            complete_on_enter=False, fail_on_enter=True, name="g1", max_duration=0.01
        )
    )
    cg.add_goal(
        MockGoal(
            complete_on_enter=False, fail_on_enter=True, name="g2", max_duration=0.01
        )
    )
    cg.enter()
    assert cg.state == GoalState.FAILED


# --- NONE_ACCOMPLISHED ---


def test_none_accomplished_all_fail():
    cg = ComplexGoal(name="cg", algorithm=ComplexGoalAlgorithm.NONE_ACCOMPLISHED)
    cg.add_goal(
        MockGoal(
            complete_on_enter=False, fail_on_enter=True, name="g1", max_duration=0.01
        )
    )
    cg.add_goal(
        MockGoal(
            complete_on_enter=False, fail_on_enter=True, name="g2", max_duration=0.01
        )
    )
    cg.enter()
    assert cg.state == GoalState.COMPLETED


def test_none_accomplished_one_completes():
    cg = ComplexGoal(name="cg", algorithm=ComplexGoalAlgorithm.NONE_ACCOMPLISHED)
    cg.add_goal(MockGoal(complete_on_enter=True, name="g1"))
    cg.add_goal(
        MockGoal(
            complete_on_enter=False, fail_on_enter=True, name="g2", max_duration=0.01
        )
    )
    cg.enter()
    assert cg.state == GoalState.FAILED


# --- AT_LEAST_ONE_ACCOMPLISHED ---


def test_at_least_one_accomplished_one_completes():
    cg = ComplexGoal(
        name="cg", algorithm=ComplexGoalAlgorithm.AT_LEAST_ONE_ACCOMPLISHED
    )
    cg.add_goal(MockGoal(complete_on_enter=True, name="g1"))
    cg.add_goal(
        MockGoal(
            complete_on_enter=False, fail_on_enter=True, name="g2", max_duration=0.01
        )
    )
    cg.enter()
    assert cg.state == GoalState.COMPLETED


def test_at_least_one_accomplished_none_complete():
    cg = ComplexGoal(
        name="cg", algorithm=ComplexGoalAlgorithm.AT_LEAST_ONE_ACCOMPLISHED
    )
    cg.add_goal(
        MockGoal(
            complete_on_enter=False, fail_on_enter=True, name="g1", max_duration=0.01
        )
    )
    cg.add_goal(
        MockGoal(
            complete_on_enter=False, fail_on_enter=True, name="g2", max_duration=0.01
        )
    )
    cg.enter()
    assert cg.state == GoalState.FAILED


# --- EXACTLY_X_ACCOMPLISHED ---


def test_exactly_x_accomplished_match():
    cg = ComplexGoal(
        name="cg",
        algorithm=ComplexGoalAlgorithm.EXACTLY_X_ACCOMPLISHED,
        accomplished=1,
    )
    cg.add_goal(MockGoal(complete_on_enter=True, name="g1"))
    cg.add_goal(
        MockGoal(
            complete_on_enter=False, fail_on_enter=True, name="g2", max_duration=0.01
        )
    )
    cg.enter()
    assert cg.state == GoalState.COMPLETED


def test_exactly_x_accomplished_no_match():
    cg = ComplexGoal(
        name="cg",
        algorithm=ComplexGoalAlgorithm.EXACTLY_X_ACCOMPLISHED,
        accomplished=2,
    )
    cg.add_goal(MockGoal(complete_on_enter=True, name="g1"))
    cg.add_goal(
        MockGoal(
            complete_on_enter=False, fail_on_enter=True, name="g2", max_duration=0.01
        )
    )
    cg.enter()
    assert cg.state == GoalState.FAILED


# --- ALL_ACCOMPLISHED_ORDERED (sequential) ---


def test_all_accomplished_ordered_all_complete():
    cg = ComplexGoal(name="cg", algorithm=ComplexGoalAlgorithm.ALL_ACCOMPLISHED_ORDERED)
    cg.add_goal(MockGoal(complete_on_enter=True, name="g1"))
    cg.add_goal(MockGoal(complete_on_enter=True, name="g2"))
    cg.enter()
    assert cg.state == GoalState.COMPLETED


def test_all_accomplished_ordered_one_fails():
    cg = ComplexGoal(name="cg", algorithm=ComplexGoalAlgorithm.ALL_ACCOMPLISHED_ORDERED)
    cg.add_goal(MockGoal(complete_on_enter=True, name="g1"))
    cg.add_goal(
        MockGoal(
            complete_on_enter=False, fail_on_enter=True, name="g2", max_duration=0.01
        )
    )
    cg.enter()
    assert cg.state == GoalState.FAILED


# --- EXACTLY_X_ACCOMPLISHED_ORDERED (sequential) ---


def test_exactly_x_ordered_match():
    cg = ComplexGoal(
        name="cg",
        algorithm=ComplexGoalAlgorithm.EXACTLY_X_ACCOMPLISHED_ORDERED,
        accomplished=1,
    )
    cg.add_goal(MockGoal(complete_on_enter=True, name="g1"))
    cg.add_goal(
        MockGoal(
            complete_on_enter=False, fail_on_enter=True, name="g2", max_duration=0.01
        )
    )
    cg.enter()
    assert cg.state == GoalState.COMPLETED


def test_exactly_x_ordered_no_match():
    cg = ComplexGoal(
        name="cg",
        algorithm=ComplexGoalAlgorithm.EXACTLY_X_ACCOMPLISHED_ORDERED,
        accomplished=3,
    )
    cg.add_goal(MockGoal(complete_on_enter=True, name="g1"))
    cg.add_goal(
        MockGoal(
            complete_on_enter=False, fail_on_enter=True, name="g2", max_duration=0.01
        )
    )
    cg.enter()
    assert cg.state == GoalState.FAILED


# --- run_seq max_duration break ---


def test_run_seq_max_duration_exceeded():
    cg = ComplexGoal(
        name="cg",
        algorithm=ComplexGoalAlgorithm.ALL_ACCOMPLISHED_ORDERED,
        max_duration=0.001,
    )
    cg.add_goal(MockGoal(complete_on_enter=True, name="g1"))
    cg.add_goal(MockGoal(complete_on_enter=True, name="g2"))
    cg.add_goal(MockGoal(complete_on_enter=True, name="g3"))
    with patch.object(cg, "get_current_elapsed", return_value=1.0):
        cg.enter()
    assert cg.state == GoalState.FAILED


# --- run_concurrent timeout ---


def test_run_concurrent_timeout():
    cg = ComplexGoal(
        name="cg",
        algorithm=ComplexGoalAlgorithm.ALL_ACCOMPLISHED,
        max_duration=0.001,
    )

    class SlowGoal(Goal):
        def on_enter(self):
            pass

        def tick(self):
            import time

            time.sleep(0.1)

    cg.add_goal(SlowGoal(name="slow1", max_duration=10.0))
    cg.enter()
    assert cg.state == GoalState.FAILED


# --- run_concurrent exception in goal ---


def test_run_concurrent_goal_exception():
    cg = ComplexGoal(name="cg", algorithm=ComplexGoalAlgorithm.ALL_ACCOMPLISHED)

    class ErrorGoal(Goal):
        def on_enter(self):
            pass

        def tick(self):
            raise RuntimeError("boom")

    cg.add_goal(ErrorGoal(name="err", max_duration=0.01))
    cg.enter()
    assert cg.state == GoalState.FAILED


# --- terminate ---


def test_terminate():
    cg = ComplexGoal(name="cg")
    g1 = MockGoal(name="g1")
    g1.set_state(GoalState.RUNNING)
    g2 = MockGoal(name="g2")
    g2.set_state(GoalState.COMPLETED)
    cg.add_goal(g1)
    cg.add_goal(g2)
    cg.terminate()
    assert g1.state == GoalState.TERMINATED
    assert g2.state == GoalState.COMPLETED
    assert cg.state == GoalState.TERMINATED


# --- terminate_all_goals ---


def test_terminate_all_goals_skips_finished():
    cg = ComplexGoal(name="cg")
    g_completed = MockGoal(name="gc")
    g_completed.set_state(GoalState.COMPLETED)
    g_failed = MockGoal(name="gf")
    g_failed.set_state(GoalState.FAILED)
    g_terminated = MockGoal(name="gt")
    g_terminated.set_state(GoalState.TERMINATED)
    g_running = MockGoal(name="gr")
    g_running.set_state(GoalState.RUNNING)
    cg.add_goal(g_completed)
    cg.add_goal(g_failed)
    cg.add_goal(g_terminated)
    cg.add_goal(g_running)
    cg.terminate_all_goals()
    assert g_completed.state == GoalState.COMPLETED
    assert g_failed.state == GoalState.FAILED
    assert g_terminated.state == GoalState.TERMINATED
    assert g_running.state == GoalState.TERMINATED


# --- _get_results_list ---


def test_get_results_list():
    cg = ComplexGoal(name="cg")
    g1 = MockGoal(name="g1")
    g1.set_state(GoalState.COMPLETED)
    g2 = MockGoal(name="g2")
    g2.set_state(GoalState.FAILED)
    g3 = MockGoal(name="g3")
    g3.set_state(GoalState.COMPLETED)
    cg.add_goal(g1)
    cg.add_goal(g2)
    cg.add_goal(g3)
    assert cg._get_results_list() == [1, 0, 1]


def test_get_results_list_empty():
    cg = ComplexGoal(name="cg")
    assert cg._get_results_list() == []


# --- enter with min_duration failure ---


def test_enter_min_duration_not_met():
    cg = ComplexGoal(
        name="cg",
        algorithm=ComplexGoalAlgorithm.ALL_ACCOMPLISHED,
        min_duration=1000.0,
    )
    cg.add_goal(MockGoal(complete_on_enter=True, name="g1"))
    cg.enter()
    assert cg.state == GoalState.FAILED


# --- enter with max_duration exceeded for non-NONE algorithms ---


def test_enter_max_duration_exceeded_all_accomplished():
    cg = ComplexGoal(
        name="cg",
        algorithm=ComplexGoalAlgorithm.ALL_ACCOMPLISHED,
        max_duration=0.001,
    )
    cg.add_goal(MockGoal(complete_on_enter=True, name="g1"))
    with patch.object(cg, "get_current_elapsed", return_value=1.0):
        cg.enter()
    assert cg.state == GoalState.FAILED


# --- enter when state is already FAILED (no further state change) ---


def test_enter_already_failed_no_change():
    cg = ComplexGoal(name="cg", algorithm=ComplexGoalAlgorithm.ALL_ACCOMPLISHED)
    cg.add_goal(
        MockGoal(
            complete_on_enter=False, fail_on_enter=True, name="g1", max_duration=0.01
        )
    )
    cg.add_goal(
        MockGoal(
            complete_on_enter=False, fail_on_enter=True, name="g2", max_duration=0.01
        )
    )
    cg.enter()
    assert cg.state == GoalState.FAILED


# --- enter returns self ---


def test_enter_returns_self():
    cg = ComplexGoal(name="cg")
    cg.add_goal(MockGoal(complete_on_enter=True, name="g1"))
    result = cg.enter()
    assert result is cg


# --- NONE_ACCOMPLISHED with max_duration exceeded does NOT fail from enter ---


def test_none_accomplished_max_duration_not_failed_by_enter():
    cg = ComplexGoal(
        name="cg",
        algorithm=ComplexGoalAlgorithm.NONE_ACCOMPLISHED,
        max_duration=0.001,
    )
    cg.add_goal(
        MockGoal(
            complete_on_enter=False, fail_on_enter=True, name="g1", max_duration=0.001
        )
    )
    cg.enter()
    assert cg.state == GoalState.COMPLETED


# --- EXACTLY_X with max_duration exceeded does NOT fail from enter ---


def test_exactly_x_max_duration_not_failed_by_enter():
    cg = ComplexGoal(
        name="cg",
        algorithm=ComplexGoalAlgorithm.EXACTLY_X_ACCOMPLISHED,
        accomplished=1,
        max_duration=0.001,
    )
    cg.add_goal(MockGoal(complete_on_enter=True, name="g1"))
    cg.enter()
    assert cg.state == GoalState.COMPLETED


# --- No goals (ThreadPoolExecutor(0) raises ValueError) ---


def test_no_goals_raises_value_error():
    cg = ComplexGoal(name="cg", algorithm=ComplexGoalAlgorithm.ALL_ACCOMPLISHED)
    with pytest.raises(ValueError):
        cg.enter()



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor, as_completed
from concurrent.futures._base import TimeoutError
from enum import IntEnum
from typing import Any

from commlib.node import Node

from goalee.goal import Goal, GoalState
from goalee.rtmonitor import RTMonitor


class ComplexGoalAlgorithm(IntEnum):
    ALL_ACCOMPLISHED = 0
    ALL_ACCOMPLISHED_ORDERED = 1
    NONE_ACCOMPLISHED = 2
    AT_LEAST_ONE_ACCOMPLISHED = 3
    EXACTLY_X_ACCOMPLISHED = 4
    EXACTLY_X_ACCOMPLISHED_ORDERED = 5


class ComplexGoal(Goal):
    def __init__(
        self,
        comm_node: Node | None = None,
        name: str | None = None,
        algorithm: ComplexGoalAlgorithm | None = None,
        accomplished: int | None = None,
        event_emitter: Any | None = None,
        max_duration: float | None = None,
        min_duration: float | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            event_emitter=event_emitter,
            name=name,
            max_duration=max_duration,
            min_duration=min_duration,
            *args,
            **kwargs,
        )
        # self.set_comm_node(comm_node)
        self._goals = []
        if algorithm is None:
            algorithm = ComplexGoalAlgorithm.ALL_ACCOMPLISHED
        self._algorithm = algorithm
        self._x_accomplished = accomplished

    @property
    def goals(self):
        return self._goals

    def set_state_cascaded(self, state: GoalState):
        for goal in self._goals:
            goal.set_state(state)
        self.set_state(state)

    def set_tick_freq(self, freq: int):
        for goal in self._goals:
            goal.set_tick_freq(freq)

    def serialize(self):
        return {
            **super().serialize(),
            "algorithm": self._algorithm.name,
            "goals": [goal.serialize() for goal in self._goals],
        }

    def enter(self, rtmonitor: RTMonitor = None):
        self.set_state(GoalState.RUNNING)
        self.on_enter()

        elapsed = self.get_current_elapsed()
        self._duration = elapsed
        if self._state == GoalState.FAILED:
            pass
        elif (
            self._min_duration not in (None, 0)
            and elapsed < self._min_duration
            or self._algorithm
            not in (
                ComplexGoalAlgorithm.NONE_ACCOMPLISHED,
                ComplexGoalAlgorithm.EXACTLY_X_ACCOMPLISHED,
                ComplexGoalAlgorithm.EXACTLY_X_ACCOMPLISHED_ORDERED,
            )
            and self._max_duration not in (None, 0)
            and elapsed > self._max_duration
        ):
            self.set_state(GoalState.FAILED)

        self.on_exit()
        return self

    def on_enter(self):
        self.log_debug(
            f"Starting ComplexGoal <{self._name}>:\n"
            f"Parameters:\n"
            f"  Algorithm: {self._algorithm.name}\n"
            f"  X-Accomplished: {self._x_accomplished}\n"
            f"  Max Duration: {self._max_duration}\n"
            f"  Min Duration: {self._min_duration}\n"
            f"Internal Goals: {[f'{g.__class__.__name__}:{g.name}' for g in self._goals]}"
        )
        self._ts_start = self.get_current_ts()

        if self._algorithm in (
            ComplexGoalAlgorithm.ALL_ACCOMPLISHED_ORDERED,
            ComplexGoalAlgorithm.EXACTLY_X_ACCOMPLISHED_ORDERED,
        ):
            self.run_seq()
        else:
            self.run_concurrent()

        self.calc_result()

        self.log_debug(
            f"Finished ComplexGoal <{self.__class__.__name__}:{self._name}>\n"
            f"  Mode {self._algorithm.name}\n"
            f"  Results: {self._get_results_list()}"
        )

    def run_seq(self):
        for g in self._goals:
            g.enter()
            if (
                self._max_duration is not None
                and self.get_current_elapsed() > self._max_duration
            ):
                self.set_state(GoalState.FAILED)
                break

    def run_concurrent(self):
        """
        Executes the 'enter' method of each goal in self._goals concurrently using a thread pool.

        This method creates a thread pool with a number of threads equal to the number of goals.
        Each goal's 'enter' method is submitted to the thread pool for execution. The results
        of the 'enter' methods are collected and returned once all threads have completed.

        If a TimeoutError occurs during the execution, it is caught and ignored.

        The thread pool is shut down after all tasks are completed or if an exception occurs.
        """
        n_threads = len(self._goals)
        futures = []
        executor = ThreadPoolExecutor(n_threads)
        for goal in self._goals:
            future = executor.submit(goal.enter)
            futures.append(future)
        try:
            for f in as_completed(futures, timeout=self._max_duration):
                try:
                    goal = f.result()
                    if (
                        goal.state == GoalState.COMPLETED
                        and self._algorithm
                        == ComplexGoalAlgorithm.AT_LEAST_ONE_ACCOMPLISHED
                    ):
                        self.log_debug(
                            f"Goal <ComplexGoal:{goal.name}> reached at least one accomplished"
                        )
                        self.terminate_all_goals()
                        break
                except Exception as e:
                    self.log_error(f"Error in goal execution: {e}")
        except TimeoutError:
            pass
        executor.shutdown(wait=False, cancel_futures=True)

    def terminate(self):
        self.terminate_all_goals()
        return super().terminate()

    def terminate_all_goals(self):
        for goal in self._goals:
            if goal.state not in (
                GoalState.COMPLETED,
                GoalState.FAILED,
                GoalState.TERMINATED,
            ):
                goal.terminate()

    def calc_result(self):
        res_list = self._get_results_list()
        self.log_info(f"<{self.__class__.__name__}:{self._name}> results: {res_list}")
        completed = res_list.count(1)
        # failed = res_list.count(0)
        self._evaluate_results(completed, res_list)

    def _get_results_list(self):
        res_list = []
        for goal in self._goals:
            if goal._state == GoalState.COMPLETED:
                res_list.append(1)
            else:
                res_list.append(0)
        return res_list

    def _evaluate_results(self, completed, res_list):
        if self._algorithm == ComplexGoalAlgorithm.ALL_ACCOMPLISHED:
            self._set_state_all_accomplished(completed, res_list)
        elif self._algorithm == ComplexGoalAlgorithm.NONE_ACCOMPLISHED:
            self._set_state_none_accomplished(completed)
        elif self._algorithm == ComplexGoalAlgorithm.AT_LEAST_ONE_ACCOMPLISHED:
            self._set_state_at_least_one_accomplished(completed)
        elif self._algorithm == ComplexGoalAlgorithm.EXACTLY_X_ACCOMPLISHED:
            self._set_state_exactly_x_accomplished(completed)
        elif self._algorithm == ComplexGoalAlgorithm.ALL_ACCOMPLISHED_ORDERED:
            self._set_state_all_accomplished(completed, res_list)
        elif self._algorithm == ComplexGoalAlgorithm.EXACTLY_X_ACCOMPLISHED_ORDERED:
            self._set_state_exactly_x_accomplished(completed)

    def _set_state_all_accomplished(self, completed, res_list):
        if completed == len(res_list):
            self.set_state(GoalState.COMPLETED)
        else:
            self.set_state(GoalState.FAILED)

    def _set_state_none_accomplished(self, completed):
        if completed == 0:
            self.set_state(GoalState.COMPLETED)
        else:
            self.set_state(GoalState.FAILED)

    def _set_state_at_least_one_accomplished(self, completed):
        if completed > 0:
            self.set_state(GoalState.COMPLETED)
        else:
            self.set_state(GoalState.FAILED)

    def _set_state_exactly_x_accomplished(self, completed):
        if completed == self._x_accomplished:
            self.set_state(GoalState.COMPLETED)
        else:
            self.set_state(GoalState.FAILED)

    def add_goal(self, goal: Goal):
        if self._max_duration is not None:
            if (
                goal._max_duration is None or goal._max_duration > self._max_duration
            ) and self._max_duration is not None:
                goal._max_duration = self._max_duration
                self.log_debug(
                    f"Goal <{goal.__class__.__name__}:{goal.name}> max duration set to {self._max_duration}"
                )
        if self._min_duration is not None:
            if (
                goal._min_duration is None or goal._min_duration < self._min_duration
            ) and self._min_duration is not None:
                goal._min_duration = self._min_duration
                self.log_debug(
                    f"Goal <{goal.__class__.__name__}:{goal.name}> min duration set to {self._min_duration}"
                )
        self._goals.append(goal)

    def reset(self):
        self.set_state(GoalState.IDLE)
        self._ts_start = -1.0
        self._ts_exit = -1.0
        self._ts_hold = -1.0
        self._duration = -1.0
        for goal in self._goals:
            goal.reset()
        self.on_reset()
